#!/usr/bin/env python
# License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>

import fcntl
import math
import os
import re
import string
import sys
from collections.abc import Callable, Generator, Iterable, Iterator, Mapping, Sequence
from contextlib import contextmanager, suppress
from functools import lru_cache
from re import Match, Pattern
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    NamedTuple,
    NoReturn,
    Optional,
    cast,
)

from .constants import (
    clear_handled_signals,
    config_dir,
    is_macos,
    is_wayland,
    kitten_exe,
    runtime_dir,
    shell_path,
    ssh_control_master_template,
)
from .fast_data_types import WINDOW_FULLSCREEN, WINDOW_HIDDEN, WINDOW_MAXIMIZED, WINDOW_MINIMIZED, WINDOW_NORMAL, Color, Shlex, get_options, monotonic, open_tty
from .fast_data_types import timed_debug_print as _timed_debug_print
from .types import run_once
from .typing_compat import AddressFamily, PopenType, StartupCtx

if TYPE_CHECKING:
    import tarfile

    from .fast_data_types import OSWindowSize
    from .options.types import Options
else:
    Options = object


class Flag:

    def __init__(self, initial_val: bool = True) -> None:
        self.val = initial_val

    def __enter__(self) -> None:
        self.val ^= True

    def __exit__(self, *a: object) -> None:
        self.val ^= True

    def __bool__(self) -> bool:
        return self.val


disallow_expand_vars = Flag(False)


def expandvars(val: str, env: Mapping[str, str] = {}, fallback_to_os_env: bool = True) -> str:
    '''
    Expand $VAR and ${VAR} Use $$ for a literal $
    '''

    def sub(m: 'Match[str]') -> str:
        key = m.group(1) or m.group(2)
        result = env.get(key)
        if result is None and fallback_to_os_env:
            result = os.environ.get(key)
        if result is None:
            result = m.group()
        return result

    if disallow_expand_vars or '$' not in val:
        return val

    return re.sub(r'\$(?:(\w+)|\{([^}]+)\})', sub, val.replace('$$', '\0')).replace('\0', '$')


@lru_cache(maxsize=2)
def sgr_sanitizer_pat(for_splitting: bool = False) -> 're.Pattern[str]':
    pat = '\033\\[.*?m'
    if for_splitting:
        return re.compile(f'({pat})')
    return re.compile(pat)


@run_once
def kitty_ansi_sanitizer_pat() -> 're.Pattern[str]':
    # removes ANSI sequences generated by kitty's ANSI output routines. Not
    # suitable for stripping general ANSI sequences
    return re.compile(r'\x1b(?:\[[0-9;:]*?m|\].*?\x1b\\)')


def platform_window_id(os_window_id: int) -> int | None:
    if is_macos:
        from .fast_data_types import cocoa_window_id
        with suppress(Exception):
            return cocoa_window_id(os_window_id)
    if not is_wayland():
        from .fast_data_types import x11_window_id
        with suppress(Exception):
            return x11_window_id(os_window_id)
    return None


def safe_print(*a: Any, **k: Any) -> None:
    with suppress(Exception):
        print(*a, **k)


def log_error(*a: Any, **k: str) -> None:
    from .fast_data_types import log_error_string
    output = getattr(log_error, 'redirect', log_error_string)
    with suppress(Exception):
        msg = k.get('sep', ' ').join(map(str, a)) + k.get('end', '')
        output(msg)


@contextmanager
def suppress_error_logging() -> Iterator[None]:
    before = getattr(log_error, 'redirect', suppress_error_logging)
    setattr(log_error, 'redirect', lambda *a: None)
    try:
        yield
    finally:
        if before is suppress_error_logging:
            delattr(log_error, 'redirect')
        else:
            setattr(log_error, 'redirect', before)


def ceil_int(x: float) -> int:
    return int(math.ceil(x))


def sanitize_title(x: str) -> str:
    return re.sub(r'\s+', ' ', re.sub(r'[\0-\x19\x80-\x9f]', '', x))


def color_as_int(val: Color) -> int:
    return int(val) & 0xffffff


def color_from_int(val: int) -> Color:
    return Color((val >> 16) & 0xFF, (val >> 8) & 0xFF, val & 0xFF)


class ScreenSize(NamedTuple):
    rows: int
    cols: int
    width: int
    height: int
    cell_width: int
    cell_height: int


def read_screen_size(fd: int = -1) -> ScreenSize:
    import array
    import fcntl
    import termios
    buf = array.array('H', [0, 0, 0, 0])
    if fd < 0:
        fd = sys.stdout.fileno()
    fcntl.ioctl(fd, termios.TIOCGWINSZ, cast(bytearray, buf))
    rows, cols, width, height = tuple(buf)
    cell_width, cell_height = width // (cols or 1), height // (rows or 1)
    return ScreenSize(rows, cols, width, height, cell_width, cell_height)


class ScreenSizeGetter:
    changed = True
    Size = ScreenSize
    ans: ScreenSize | None = None

    def __init__(self, fd: int | None):
        if fd is None:
            fd = sys.stdout.fileno()
        self.fd = fd

    def __call__(self) -> ScreenSize:
        if self.changed:
            self.ans = read_screen_size(self.fd)
            self.changed = False
        return cast(ScreenSize, self.ans)


@lru_cache(maxsize=64, typed=True)
def screen_size_function(fd: int | None = None) -> ScreenSizeGetter:
    return ScreenSizeGetter(fd)


def fit_image(width: int, height: int, pwidth: int, pheight: int) -> tuple[int, int]:
    from math import floor
    if height > pheight:
        corrf = pheight / float(height)
        width, height = floor(corrf * width), pheight
    if width > pwidth:
        corrf = pwidth / float(width)
        width, height = pwidth, floor(corrf * height)
    if height > pheight:
        corrf = pheight / float(height)
        width, height = floor(corrf * width), pheight

    return int(width), int(height)


def base64_encode(
    integer: int,
    chars: str = string.ascii_uppercase + string.ascii_lowercase + string.digits +
    '+/'
) -> str:
    ans = ''
    while True:
        integer, remainder = divmod(integer, 64)
        ans = chars[remainder] + ans
        if integer == 0:
            break
    return ans


def command_for_open(program: str | list[str] = 'default') -> list[str]:
    if isinstance(program, str):
        from .conf.utils import to_cmdline
        program = to_cmdline(program)
    if program == ['default']:
        cmd = ['open'] if is_macos else ['xdg-open']
    else:
        cmd = program
    return cmd


def open_cmd(cmd: Iterable[str] | list[str], arg: None | Iterable[str] | str = None,
             cwd: str | None = None, extra_env: dict[str, str] | None = None) -> 'PopenType[bytes]':
    import subprocess
    if arg is not None:
        cmd = list(cmd)
        if isinstance(arg, str):
            cmd.append(arg)
        else:
            cmd.extend(arg)
    env: dict[str, str] | None = None
    if extra_env:
        env = os.environ.copy()
        env.update(extra_env)
    return subprocess.Popen(
        tuple(cmd), stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, cwd=cwd or None,
        preexec_fn=clear_handled_signals, env=env)


def open_url(url: str, program: str | list[str] = 'default', cwd: str | None = None, extra_env: dict[str, str] | None = None) -> 'PopenType[bytes]':
    return open_cmd(command_for_open(program), url, cwd=cwd, extra_env=extra_env)


def init_startup_notification_x11(window_handle: int, startup_id: str | None = None) -> Optional['StartupCtx']:
    # https://specifications.freedesktop.org/startup-notification-spec/startup-notification-latest.txt
    from kitty.fast_data_types import init_x11_startup_notification
    sid = startup_id or os.environ.pop('DESKTOP_STARTUP_ID', None)  # ensure child processes don't get this env var
    if not sid:
        return None
    from .fast_data_types import x11_display
    display = x11_display()
    if not display:
        return None
    return init_x11_startup_notification(display, window_handle, sid)


def end_startup_notification_x11(ctx: 'StartupCtx') -> None:
    from kitty.fast_data_types import end_x11_startup_notification
    end_x11_startup_notification(ctx)


def init_startup_notification(window_handle: int | None, startup_id: str | None = None) -> Optional['StartupCtx']:
    if is_macos or is_wayland():
        return None
    if window_handle is None:
        log_error('Could not perform startup notification as window handle not present')
        return None
    try:
        try:
            return init_startup_notification_x11(window_handle, startup_id)
        except OSError as e:
            if not str(e).startswith("Failed to load libstartup-notification"):
                raise e
            log_error(
                f'{e}. This has two main effects:',
                'There will be no startup feedback and when using --single-instance, kitty windows may start on an incorrect desktop/workspace.')
    except Exception:
        import traceback
        traceback.print_exc()
    return None


def end_startup_notification(ctx: Optional['StartupCtx']) -> None:
    if not ctx:
        return
    if is_macos or is_wayland():
        return
    try:
        end_startup_notification_x11(ctx)
    except Exception:
        import traceback
        traceback.print_exc()


class startup_notification_handler:

    # WARNING: This only works on X11 on other platforms extra_callback will be called
    # after the window is shown, not before, as they do not do two stage window
    # creation.

    def __init__(self, do_notify: bool = True, startup_id: str | None = None, extra_callback: Callable[[int], None] | None = None):
        self.do_notify = do_notify
        self.startup_id = startup_id
        self.extra_callback = extra_callback
        self.ctx: Optional['StartupCtx'] = None

    def __enter__(self) -> Callable[[int], None]:

        def pre_show_callback(window_handle: int) -> None:
            if self.extra_callback is not None:
                self.extra_callback(window_handle)
            if self.do_notify:
                self.ctx = init_startup_notification(window_handle, self.startup_id)

        return pre_show_callback

    def __exit__(self, *a: Any) -> None:
        if self.ctx is not None:
            end_startup_notification(self.ctx)


def unix_socket_directories() -> Iterator[str]:
    import tempfile
    home = os.path.expanduser('~')
    candidates = [tempfile.gettempdir(), home]
    if is_macos:
        from .fast_data_types import user_cache_dir
        candidates = [user_cache_dir(), '/Library/Caches']
    else:
        if os.environ.get('XDG_RUNTIME_DIR'):
            candidates.insert(0, os.environ['XDG_RUNTIME_DIR'])
    for loc in candidates:
        if os.access(loc, os.W_OK | os.R_OK | os.X_OK):
            yield loc


def unix_socket_paths(name: str, ext: str = '.lock') -> Generator[str, None, None]:
    home = os.path.expanduser('~')
    for loc in unix_socket_directories():
        filename = ('.' if loc == home else '') + name + ext
        yield os.path.join(loc, filename)


def parse_address_spec(spec: str) -> tuple[AddressFamily, tuple[str, int] | str, str | None]:
    import socket
    try:
        protocol, rest = spec.split(':', 1)
    except ValueError:
        raise ValueError(f'Invalid listen-on value: {spec} must be of the form protocol:address')
    socket_path = None
    address: str | tuple[str, int] = ''
    if protocol == 'unix':
        family = socket.AF_UNIX
        address = rest
        if address.startswith('@') and len(address) > 1:
            address = '\0' + address[1:]
        else:
            socket_path = address
    elif protocol in ('tcp', 'tcp6'):
        family = socket.AF_INET if protocol == 'tcp' else socket.AF_INET6
        if rest.startswith('['):  # ]
            host = rest[1:]
            host, sep, leftover = host.rpartition(']')
            _, port = leftover.rsplit(':', 1)
            if ':' in host and protocol == 'tcp':
                family = socket.AF_INET6
        else:
            host, port = rest.rsplit(':', 1)
        address = host, int(port)
    else:
        raise ValueError(f'Unknown protocol in listen-on value: {spec}')
    return family, address, socket_path


def parse_os_window_state(state: str) -> int:
    match state:
        case 'normal':
            return WINDOW_NORMAL
        case 'maximized':
            return WINDOW_MAXIMIZED
        case 'minimized':
            return WINDOW_MINIMIZED
        case 'fullscreen' | 'fullscreened':
            return WINDOW_FULLSCREEN
        case 'hidden':
            return WINDOW_HIDDEN
        case _:
            return WINDOW_NORMAL


def write_all(fd: int, data: str | bytes, block_until_written: bool = True) -> None:
    if isinstance(data, str):
        data = data.encode('utf-8')
    mvd = memoryview(data)
    while len(mvd) > 0:
        try:
            n = os.write(fd, mvd)
        except BlockingIOError:
            if not block_until_written:
                raise
            continue
        if not n:
            break
        mvd = mvd[n:]


class TTYIO:

    def __init__(self, read_with_timeout: bool = True):
        self.read_with_timeout = read_with_timeout

    def __enter__(self) -> 'TTYIO':
        self.tty_fd, self.original_termios = open_tty(self.read_with_timeout)
        return self

    def __exit__(self, *a: Any) -> None:
        from .fast_data_types import close_tty
        close_tty(self.tty_fd, self.original_termios)

    def wait_till_read_available(self) -> bool:
        if self.read_with_timeout:
            raise ValueError('Cannot wait when TTY is set to read with timeout')
        import select
        rd = select.select([self.tty_fd], [], [])[0]
        return bool(rd)

    def read(self, limit: int) -> bytes:
        return os.read(self.tty_fd, limit)

    def send(self, data: str | bytes | Iterable[str | bytes]) -> None:
        if isinstance(data, (str, bytes)):
            write_all(self.tty_fd, data)
        else:
            for chunk in data:
                write_all(self.tty_fd, chunk)

    def recv(self, more_needed: Callable[[bytes], bool], timeout: float, sz: int = 1) -> None:
        fd = self.tty_fd
        start_time = monotonic()
        while timeout > monotonic() - start_time:
            # will block for 0.1 secs waiting for data because we have set
            # VMIN=0 VTIME=1 in termios
            data = os.read(fd, sz)
            if data and not more_needed(data):
                break


def set_echo(fd: int = -1, on: bool = False) -> tuple[int, list[int | list[bytes | int]]]:
    import termios
    if fd < 0:
        fd = sys.stdin.fileno()
    old = termios.tcgetattr(fd)
    new = termios.tcgetattr(fd)
    if on:
        new[3] |= termios.ECHO
    else:
        new[3] &= ~termios.ECHO
    termios.tcsetattr(fd, termios.TCSADRAIN, new)
    return fd, old


@contextmanager
def no_echo(fd: int = -1) -> Iterator[None]:
    import termios
    fd, old = set_echo(fd)
    try:
        yield
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old)


def natsort_ints(iterable: Iterable[str]) -> list[str]:

    def convert(text: str) -> int | str:
        return int(text) if text.isdigit() else text

    def alphanum_key(key: str) -> tuple[int | str, ...]:
        return tuple(map(convert, re.split(r'(\d+)', key)))

    return sorted(iterable, key=alphanum_key)


def get_hostname(fallback: str = '') -> str:
    import socket
    try:
        return socket.gethostname() or fallback
    except Exception:
        return fallback


def resolve_editor_cmd(editor: str, shell_env: Mapping[str, str]) -> str | None:
    import shlex
    editor_cmd = list(shlex_split(editor))
    editor_exe = (editor_cmd or ('',))[0]
    if editor_exe and os.path.isabs(editor_exe):
        return editor
    if not editor_exe:
        return None

    def patched(exe: str) -> str:
        editor_cmd[0] = exe
        return ' '.join(map(shlex.quote, editor_cmd))

    if shell_env is os.environ:
        q = which(editor_exe, only_system=True)
        if q:
            return patched(q)
    elif 'PATH' in shell_env:
        import shutil
        q = shutil.which(editor_exe, path=shell_env['PATH'])
        if q:
            return patched(q)
    return None


def get_editor_from_env(env: Mapping[str, str]) -> str | None:
    for var in ('VISUAL', 'EDITOR'):
        editor = env.get(var)
        if editor:
            editor = resolve_editor_cmd(editor, env)
            if editor:
                return editor
    return None


def get_editor_from_env_vars(opts: Options | None = None) -> list[str]:
    from .child import default_env
    editor = get_editor_from_env(default_env())
    if not editor:
        shell_env = read_shell_environment(opts)
        editor = get_editor_from_env(shell_env)

    for ans in (editor, 'vim', 'nvim', 'vi', 'emacs', 'hx', 'kak', 'micro', 'nano', 'vis'):
        if ans and which(next(shlex_split(ans)), only_system=True):
            break
    else:
        ans = 'vim'
    return list(shlex_split(ans))


def get_editor(opts: Options | None = None, path_to_edit: str = '', line_number: int = 0) -> list[str]:
    if opts is None:
        try:
            opts = get_options()
        except RuntimeError:
            # we are in a kitten
            from .cli import create_default_opts
            opts = create_default_opts()
    if opts.editor == '.':
        ans = get_editor_from_env_vars()
    else:
        ans = list(shlex_split(opts.editor))
    ans[0] = os.path.expanduser(ans[0])
    if path_to_edit:
        if line_number:
            eq = os.path.basename(ans[0]).lower()
            if eq in ('code', 'code.exe'):
                path_to_edit += f':{line_number}'
                ans.append('--goto')
            else:
                ans.append(f'+{line_number}')
        ans.append(path_to_edit)
    return ans


def edit_file(path: str = '') -> NoReturn:
    ' This exists for: map whatever launch kitty +runpy "from kitty.utils import *; edit_file()" to edit kitty config '
    from .config import prepare_config_file_for_editing
    editor = get_editor()
    path = path or prepare_config_file_for_editing()
    editor.append(path)
    os.execlp(editor[0], *editor)


def is_path_in_temp_dir(path: str) -> bool:
    if not path:
        return False

    def abspath(x: str | None) -> str:
        if x:
            x = os.path.abspath(os.path.realpath(x))
        return x or ''

    import tempfile
    path = abspath(path)
    candidates = frozenset(map(abspath, ('/tmp', '/dev/shm', os.environ.get('TMPDIR', None), tempfile.gettempdir())))
    for q in candidates:
        if q and path.startswith(q):
            return True
    return False


def is_ok_to_read_image_file(path: str, fd: int) -> bool:
    import stat
    path = os.path.abspath(os.path.realpath(path))
    try:
        path_stat = os.stat(path, follow_symlinks=True)
        fd_stat = os.fstat(fd)
    except OSError:
        return False
    if not os.path.samestat(path_stat, fd_stat):
        return False
    parts = path.split(os.sep)[1:]
    if len(parts) < 1:
        return False
    if parts[0] in ('sys', 'proc', 'dev'):
        if parts[0] == 'dev':
            return len(parts) > 2 and parts[1] == 'shm'
        return False
    return stat.S_ISREG(fd_stat.st_mode)


def resolve_abs_or_config_path(path: str, env: Mapping[str, str] | None = None, conf_dir: str | None = None) -> str:
    path = os.path.expanduser(path)
    path = expandvars(path, env or {})
    if not os.path.isabs(path):
        path = os.path.join(conf_dir or config_dir, path)
    return path


def resolve_custom_file(path: str) -> str:
    opts: Options | None = None
    with suppress(RuntimeError):
        opts = get_options()
    return resolve_abs_or_config_path(path, opts.env if opts else {})


def func_name(f: Any) -> str:
    if hasattr(f, '__name__'):
        return str(f.__name__)
    if hasattr(f, 'func') and hasattr(f.func, '__name__'):
        return str(f.func.__name__)
    return str(f)


def resolved_shell(opts: Options | None = None) -> list[str]:
    q: str = getattr(opts, 'shell', '.')
    if q == '.':
        ans = [shell_path]
    else:
        env = {}
        if opts is not None:
            env['TERM'] = opts.term
        if 'SHELL' not in os.environ:
            env['SHELL'] = shell_path
        if 'HOME' not in os.environ:
            env['HOME'] = os.path.expanduser('~')
        if 'USER' not in os.environ:
            import pwd
            env['USER'] = pwd.getpwuid(os.geteuid()).pw_name
        def expand(x: str) -> str:
            return expandvars(x, env)
        ans = list(map(expand, shlex_split(q)))
    return ans


@run_once
def system_paths_on_macos() -> tuple[str, ...]:
    entries, seen = [], set()

    def add_from_file(x: str) -> None:
        try:
            f = open(x)
        except (FileNotFoundError, PermissionError):
            return
        with f:
            for line in f:
                line = line.strip()
                if line and not line.startswith('#') and line not in seen:
                    if os.path.isdir(line):
                        seen.add(line)
                        entries.append(line)
    try:
        files = os.listdir('/etc/paths.d')
    except (FileNotFoundError, PermissionError):
        files = []
    for name in sorted(files):
        add_from_file(os.path.join('/etc/paths.d', name))
    add_from_file('/etc/paths')
    return tuple(entries)


def which(name: str, only_system: bool = False) -> str | None:
    if os.sep in name:
        return name
    import shutil

    opts: Options | None = None
    with suppress(RuntimeError):
        opts = get_options()

    tried_paths = set()
    paths = []
    append_paths = []
    if opts and opts.exe_search_path:
        for x in opts.exe_search_path:
            x = x.strip()
            if x:
                if x[0] == '-':
                    tried_paths.add(os.path.expanduser(x[1:]))
                elif x[0] == '+':
                    append_paths.append(os.path.expanduser(x[1:]))
                else:
                    paths.append(os.path.expanduser(x))
    ep = os.environ.get('PATH')
    if ep:
        paths.extend(ep.split(os.pathsep))
    paths.append(os.path.expanduser('~/.local/bin'))
    paths.append(os.path.expanduser('~/bin'))
    paths.extend(append_paths)
    ans = shutil.which(name, path=os.pathsep.join(x for x in paths if x not in tried_paths))
    if ans:
        return ans
    # In case PATH is messed up try a default set of paths
    if is_macos:
        system_paths = system_paths_on_macos()
    else:
        system_paths = ('/usr/local/bin', '/opt/bin', '/usr/bin', '/bin', '/usr/sbin', '/sbin')
    tried_paths |= set(paths)
    system_paths = tuple(x for x in system_paths if x not in tried_paths)
    if system_paths:
        ans = shutil.which(name, path=os.pathsep.join(system_paths))
        if ans:
            return ans
        tried_paths |= set(system_paths)
    if only_system or opts is None:
        return None
    shell_env = read_shell_environment(opts)
    for xenv in (shell_env, opts.env):
        q = xenv.get('PATH')
        if q:
            paths = [x for x in xenv['PATH'].split(os.pathsep) if x not in tried_paths]
            ans = shutil.which(name, path=os.pathsep.join(paths))
            if ans:
                return ans
            tried_paths |= set(paths)
    return None


@lru_cache(4)
def read_resolved_shell_environment(shell: tuple[str, ...]) -> MappingProxyType[str, str]:
    import subprocess
    cmdline = list(shell)
    if '-l' not in cmdline and '--login' not in cmdline:
        cmdline += ['-l']
    if '-i' not in cmdline and '--interactive' not in cmdline:
        cmdline += ['-i']
    q = os.path.basename(cmdline[0]).lower()
    has_builtin = q in ('bash', 'zsh')
    cmd = 'builtin command env -0' if has_builtin else 'command env -0'
    ans: MappingProxyType[str, str] = MappingProxyType({})

    from .child import openpty
    master, slave = openpty()
    os.set_blocking(master, False)
    try:
        p = subprocess.Popen(
            cmdline + ['-c', cmd], stdout=slave, stdin=slave, stderr=slave, start_new_session=True, close_fds=True,
            preexec_fn=clear_handled_signals)
    except FileNotFoundError:
        log_error(f'Could not find shell {cmdline[0]} to read environment')
        return ans
    with os.fdopen(master, 'rb') as stdout, os.fdopen(slave, 'wb'):
        raw = b''
        from time import monotonic
        start_time = monotonic()
        ret: int | None = None
        while monotonic() - start_time < 1.5:
            try:
                ret = p.wait(0.01)
            except subprocess.TimeoutExpired:
                ret = None
            with suppress(Exception):
                raw += stdout.read()
            if ret is not None:
                break
        if ret is None:
            log_error(f'Timed out waiting for shell {cmdline} to quit while reading shell environment')
            p.kill()
        elif ret == 0:
            while True:
                try:
                    x = stdout.read()
                except Exception:
                    break
                if not x:
                    break
                raw += x
            draw = raw.decode('utf-8', 'replace')
            env = {}
            for line in draw.split('\0'):
                k, sep, v = line.partition('=')
                if k and v and sep:
                    env[k] = v
            ans = MappingProxyType(env)
        else:
            log_error(f'Failed to run shell {cmdline} to read its environment')
    return ans


def read_shell_environment(opts: Options | None = None) -> MappingProxyType[str, str]:
    shell = resolved_shell(opts)
    return read_resolved_shell_environment(tuple(shell))


def parse_uri_list(text: str) -> Generator[str, None, None]:
    ' Get paths from file:// URLs '
    from urllib.parse import unquote, urlparse
    for line in text.splitlines():
        if not line or line.startswith('#'):
            continue
        if not line.startswith('file://'):
            yield line
            continue
        try:
            purl = urlparse(line, allow_fragments=False)
        except Exception:
            yield line
            continue
        if purl.path:
            yield unquote(purl.path)


def edit_config_file() -> None:
    from kitty.config import prepare_config_file_for_editing
    p = prepare_config_file_for_editing()
    editor = get_editor()
    os.execvp(editor[0], editor + [p])


class SSHConnectionData(NamedTuple):
    binary: str
    hostname: str
    port: int | None = None
    identity_file: str = ''
    extra_args: tuple[tuple[str, str], ...] = ()


def get_new_os_window_size(
    metrics: 'OSWindowSize', width: int, height: int, unit: str, incremental: bool = False, has_window_scaling: bool = True
) -> tuple[int, int]:
    if unit == 'cells':
        cw = metrics['cell_width']
        ch = metrics['cell_height']
        width *= cw
        height *= ch
        if has_window_scaling:
            width = round(width / metrics['xscale'])
            height = round(height / metrics['yscale'])
    if incremental:
        w = metrics['width'] + width
        h = metrics['height'] + height
    else:
        w = width or metrics['width']
        h = height or metrics['height']
    return w, h


def get_all_processes() -> Iterable[int]:
    if is_macos:
        from kitty.fast_data_types import get_all_processes as f
        yield from f()
    else:
        for c in os.listdir('/proc'):
            if c.isdigit():
                yield int(c)


def is_kitty_gui_cmdline(*cmd: str) -> bool:
    if not cmd:
        return False
    if os.path.basename(cmd[0]) != 'kitty':
        return False
    if len(cmd) == 1:
        return True
    s = cmd[1][:1]
    if s == '@':
        return False
    if s == '+':
        if cmd[1] == '+':
            return len(cmd) > 2 and cmd[2] == 'open'
        return cmd[1] == '+open'
    return True


def reload_conf_in_all_kitties() -> None:
    import signal

    from kitty.child import cmdline_of_pid

    for pid in get_all_processes():
        try:
            cmd = cmdline_of_pid(pid)
        except Exception:
            continue
        if cmd and is_kitty_gui_cmdline(*cmd):
            os.kill(pid, signal.SIGUSR1)


@run_once
def control_codes_pat() -> 'Pattern[str]':
    return re.compile('[\x00-\x09\x0b-\x1f\x7f-\x9f]')


def sanitize_control_codes(text: str, replace_with: str = '') -> str:
    return control_codes_pat().sub(replace_with, text)


def hold_till_enter() -> None:
    import subprocess

    from .constants import kitten_exe
    subprocess.Popen([kitten_exe(), '__hold_till_enter__']).wait()


def cleanup_ssh_control_masters() -> None:
    import glob
    import subprocess
    try:
        files = frozenset(glob.glob(os.path.join(runtime_dir(), ssh_control_master_template.format(
            kitty_pid=os.getpid(), ssh_placeholder='*'))))
    except OSError:
        return
    workers = tuple(subprocess.Popen([
        'ssh', '-o', f'ControlPath={x}', '-O', 'exit', 'kitty-unused-host-name'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
        preexec_fn=clear_handled_signals) for x in files)
    for w in workers:
        w.wait()
    for x in files:
        with suppress(OSError):
            os.remove(x)


def path_from_osc7_url(url: str | bytes) -> str:
    if isinstance(url, bytes):
        url = url.decode('utf-8')
    if url.startswith('kitty-shell-cwd://'):
        return '/' + url.split('/', 3)[-1]
    if url.startswith('file://'):
        from urllib.parse import unquote, urlparse
        return unquote(urlparse(url).path)
    return ''


@run_once
def macos_version() -> tuple[int, ...]:
    # platform.mac_ver does not work thanks to Apple's stupid "hardening", so just use sw_vers
    import subprocess
    try:
        o = subprocess.check_output(['sw_vers', '-productVersion'], stderr=subprocess.STDOUT).decode()
    except Exception:
        return 0, 0, 0
    return tuple(map(int, o.strip().split('.')))


@lru_cache(maxsize=2)
def less_version(less_exe: str = 'less') -> int:
    import subprocess
    o = subprocess.check_output([less_exe, '-V'], stderr=subprocess.STDOUT).decode()
    m = re.match(r'less (\d+)', o)
    if m is None:
        raise ValueError(f'Invalid version string for less: {o}')
    return int(m.group(1))


def is_pid_alive(pid: int) -> bool:
    try:
        os.kill(pid, 0)
    except ProcessLookupError:
        return False
    except Exception:
        pass
    return True


def safer_fork() -> int:
    pid = os.fork()
    if pid:
        # master
        import ssl
        ssl.RAND_add(os.urandom(32), 0.0)
    else:
        # child
        import atexit
        atexit._clear()
    return pid


def docs_url(which: str = '', local_docs_root: str | None = '') -> str:
    from urllib.parse import quote

    from .conf.types import resolve_ref
    from .constants import local_docs, website_url
    if local_docs_root is None:
        ld = ''
    else:
        ld = local_docs_root or local_docs()
    base, frag = which.partition('#')[::2]
    base = base.strip('/')
    if frag.startswith('ref='):
        ref = frag[4:]
        which = resolve_ref(ref, lambda x: x)
        if which.startswith('https://') or which.startswith('http://'):
            return which
        base, frag = which.partition('#')[::2]
        base = base.strip('/')
    if ld:
        base = base or 'index'
        url = f'file://{ld}/' + quote(base) + '.html'
    else:
        url = website_url(base)
    if frag:
        url += '#' + frag
    return url


def sanitize_for_bracketed_paste(text: bytes) -> bytes:
    pat = re.compile(b'(?:(?:\033\\\x5b)|(?:\x9b))201~')
    while True:
        new_text = pat.sub(b'', text)
        if new_text == text:
            break
        text = new_text
    return text


@lru_cache(maxsize=64)
def sanitize_url_for_display_to_user(url: str) -> str:
    from urllib.parse import unquote, urlparse, urlunparse
    try:
        purl = urlparse(url)
        if purl.netloc:
            purl = purl._replace(netloc=purl.netloc.encode('idna').decode('ascii'))
        if purl.path:
            purl = purl._replace(path=unquote(purl.path))
        url = urlunparse(purl)
    except Exception as e:
        log_error(e)
        url = 'Unparseable URL: ' + url
    return url


def extract_all_from_tarfile_safely(tf: 'tarfile.TarFile', dest: str) -> None:
    # Ensure that all extracted items are within dest

    def is_within_directory(directory: str, target: str) -> bool:
        abs_directory = os.path.abspath(directory)
        abs_target = os.path.abspath(target)
        prefix = os.path.commonprefix((abs_directory, abs_target))
        return prefix == abs_directory

    def safe_extract(tar: 'tarfile.TarFile', path: str = ".", numeric_owner: bool = False) -> None:
        for member in tar.getmembers():
            member_path = os.path.join(path, member.name)
            if not is_within_directory(path, member_path):
                raise ValueError(f'Attempted path traversal in tar file: {member.name}')
        tar.extractall(path, tar.getmembers(), numeric_owner=numeric_owner)

    safe_extract(tf, dest)


def is_png(path: str) -> bool:
    if path:
        with suppress(Exception), open(path, 'rb') as f:
            header = f.read(8)
            return header.startswith(b'\211PNG\r\n\032\n')
    return False


def cmdline_for_hold(cmd: Sequence[str] = (), opts: Optional['Options'] = None) -> list[str]:
    if opts is None:
        with suppress(RuntimeError):
            opts = get_options()
    if opts is None:
        from .options.types import defaults
        opts = defaults
    ksi = ' '.join(opts.shell_integration)
    import shlex
    shell = shlex.join(resolved_shell(opts))
    return [kitten_exe(), 'run-shell', f'--shell={shell}', f'--shell-integration={ksi}', '--env=KITTY_HOLD=1'] + list(cmd)


def safe_mtime(path: str) -> float | None:
    with suppress(OSError):
        return os.path.getmtime(path)
    return None


@run_once
def get_custom_window_icon() -> tuple[float, str] | tuple[None, None]:
    filenames = ['kitty.app.png']
    if is_macos:
        # On macOS, prefer icns to png.
        filenames.insert(0, 'kitty.app.icns')
    for name in filenames:
        custom_icon_path = os.path.join(config_dir, name)
        custom_icon_mtime = safe_mtime(custom_icon_path)
        if custom_icon_mtime is not None:
            return custom_icon_mtime, custom_icon_path
    return None, None


def key_val_matcher(items: Iterable[tuple[str, str]], key_pat: 're.Pattern[str]', val_pat: Optional['re.Pattern[str]']) -> bool:
    for key, val in items:
        if key_pat.search(key) is not None and (
                val_pat is None or val_pat.search(val) is not None):
            return True
    return False


def shlex_split(text: str, allow_ansi_quoted_strings: bool = False) -> Iterator[str]:
    yield from Shlex(text, allow_ansi_quoted_strings)


def shlex_split_with_positions(text: str, allow_ansi_quoted_strings: bool = False) -> Iterator[tuple[int, str]]:
    s = Shlex(text, allow_ansi_quoted_strings)
    while (q := s.next_word())[0] > -1:
        yield q


def timed_debug_print(*a: Any, sep: str = ' ', end: str = '\n') -> None:
    _timed_debug_print(sep.join(map(str, a)) + end)


def lock_file(f: BinaryIO) -> None:
    if not f.writable():
        raise ValueError('Cannot lock files not opened in writable mode')
    fcntl.lockf(f, fcntl.LOCK_EX)


def unlock_file(f: BinaryIO) -> None:
    if not f.writable():
        raise ValueError('Cannot unlock files not opened in writable mode')
    fcntl.lockf(f, fcntl.LOCK_UN)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>

from typing import TYPE_CHECKING

from .base import MATCH_WINDOW_OPTION, ArgsType, Boss, PayloadGetType, PayloadType, RCOptions, RemoteCommand, ResponseType, Window

if TYPE_CHECKING:
    from kitty.cli_stub import SetUserVarsRCOptions as CLIOptions


class SetUserVars(RemoteCommand):

    protocol_spec = __doc__ = '''
    var/list.str: List of user variables of the form NAME=VALUE
    match/str: Which windows to change the title in
    '''

    short_desc = 'Set user variables on a window'
    desc = (
        'Set user variables for the specified windows. If you use the :option:`kitten @ set-user-vars --match` option'
        ' the variables will be set for all matched windows. By default, only the window'
        ' in which the command is run is affected. If you do not specify any variables, the'
        ' current variables are printed out, one per line. To unset a variable specify just its name.'
    )
    options_spec = MATCH_WINDOW_OPTION
    args = RemoteCommand.Args(json_field='var', spec='[NAME=VALUE ...]')

    def message_to_kitty(self, global_opts: RCOptions, opts: 'CLIOptions', args: ArgsType) -> PayloadType:
        return {'match': opts.match, 'var': args, 'self': True}

    def response_from_kitty(self, boss: Boss, window: Window | None, payload_get: PayloadGetType) -> ResponseType:
        val = {}
        for x in payload_get('var') or ():
            a, sep, b = x.partition('=')
            if sep:
                val[a] = b
            else:
                val[a] = None
        lines = []
        for window in self.windows_for_match_payload(boss, window, payload_get):
            if window:
                if val:
                    for k, v in val.items():
                        window.set_user_var(k, v)
                else:
                    lines.append('\n'.join(f'{k}={v}' for k, v in window.user_vars.items()))
        return '\n\n'.join(lines)


set_user_vars = SetUserVars()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>

from collections.abc import Generator, Iterable, Iterator, Sequence
from functools import partial
from itertools import repeat
from typing import Any, Callable, NamedTuple

from kitty.borders import BorderColor
from kitty.fast_data_types import Region, get_options, set_active_window, viewport_for_window
from kitty.options.types import Options
from kitty.types import Edges, NeighborsMap, WindowGeometry, WindowMapper
from kitty.typing_compat import WindowType
from kitty.window_list import WindowGroup, WindowList


class BorderLine(NamedTuple):
    edges: Edges = Edges()
    color: BorderColor = BorderColor.inactive


class LayoutOpts:

    def __init__(self, data: dict[str, str]):
        pass

    def serialized(self) -> dict[str, Any]:
        return {}


class LayoutData(NamedTuple):
    content_pos: int = 0
    cells_per_window: int = 0
    space_before: int = 0
    space_after: int = 0
    content_size: int = 0


DecorationPairs = Sequence[tuple[int, int]]
LayoutDimension = Generator[LayoutData, None, None]
ListOfWindows = list[WindowType]


class LayoutGlobalData:
    draw_minimal_borders: bool = True
    draw_active_borders: bool = True
    alignment_x: int = 0
    alignment_y: int = 0

    central: Region = Region((0, 0, 199, 199, 200, 200))
    cell_width: int = 20
    cell_height: int = 20


lgd = LayoutGlobalData()


def idx_for_id(win_id: int, windows: Iterable[WindowType]) -> int | None:
    for i, w in enumerate(windows):
        if w.id == win_id:
            return i
    return None


def effective_draw_minimal_borders(opts: Options, has_more_than_one_visible_group: bool = True) -> bool:
    ans = opts.draw_minimal_borders and sum(opts.window_margin_width) == 0
    if not has_more_than_one_visible_group and opts.draw_window_borders_for_single_window:
        ans = False
    return ans


def set_layout_options(opts: Options) -> None:
    lgd.draw_minimal_borders = effective_draw_minimal_borders(opts)
    lgd.draw_active_borders = opts.active_border_color is not None
    lgd.alignment_x = -1 if opts.placement_strategy.endswith('left') else 1 if opts.placement_strategy.endswith('right') else 0
    lgd.alignment_y = -1 if opts.placement_strategy.startswith('top') else 1 if opts.placement_strategy.startswith('bottom') else 0


def convert_bias_map(bias: dict[int, float], number_of_windows: int, number_of_cells: int) -> Sequence[float]:
    cells_per_window, extra = divmod(number_of_cells, number_of_windows)
    cell_map = list(repeat(cells_per_window, number_of_windows))
    cell_map[-1] += extra
    base_bias = [x / number_of_cells for x in cell_map]
    return distribute_indexed_bias(base_bias, bias)


def calculate_cells_map(
    bias: None | Sequence[float] | dict[int, float],
    number_of_windows: int, number_of_cells: int
) -> list[int]:
    if isinstance(bias, dict):
        bias = convert_bias_map(bias, number_of_windows, number_of_cells)
    cells_per_window = number_of_cells // number_of_windows
    if bias is not None and number_of_windows > 1 and number_of_windows == len(bias) and cells_per_window > 5:
        cells_map = [int(b * number_of_cells) for b in bias]
        while min(cells_map) < 5:
            maxi, mini = map(cells_map.index, (max(cells_map), min(cells_map)))
            if maxi == mini:
                break
            cells_map[mini] += 1
            cells_map[maxi] -= 1
    else:
        cells_map = list(repeat(cells_per_window, number_of_windows))
    extra = number_of_cells - sum(cells_map)
    if extra > 0:
        cells_map[-1] += extra
    return cells_map


def layout_dimension(
    start_at: int, length: int, cell_length: int,
    decoration_pairs: DecorationPairs,
    alignment: int = 0,
    bias: None | Sequence[float] | dict[int, float] = None
) -> LayoutDimension:
    number_of_windows = len(decoration_pairs)
    number_of_cells = length // cell_length
    dec_vals: Iterable[int] = map(sum, decoration_pairs)
    space_needed_for_decorations = sum(dec_vals)
    extra = length - number_of_cells * cell_length
    while extra < space_needed_for_decorations:
        number_of_cells -= 1
        extra = length - number_of_cells * cell_length
    cells_map = calculate_cells_map(bias, number_of_windows, number_of_cells)
    assert sum(cells_map) == number_of_cells

    extra = length - number_of_cells * cell_length - space_needed_for_decorations
    pos = start_at  # start
    if alignment > 0:  # end
        pos += extra
    elif alignment == 0:  # center
        pos += extra // 2
    last_i = len(cells_map) - 1

    for i, cells_per_window in enumerate(cells_map):
        before_dec, after_dec = decoration_pairs[i]
        pos += before_dec
        if i == 0:
            before_space = pos - start_at
        else:
            before_space = before_dec
        content_size = cells_per_window * cell_length
        if i == last_i:
            after_space = (start_at + length) - (pos + content_size)
        else:
            after_space = after_dec
        yield LayoutData(pos, cells_per_window, before_space, after_space, content_size)
        pos += content_size + after_space


class Rect(NamedTuple):
    left: int
    top: int
    right: int
    bottom: int


def blank_rects_for_window(wg: WindowGeometry) -> Generator[Rect, None, None]:
    left_width, right_width = wg.spaces.left, wg.spaces.right
    top_height, bottom_height = wg.spaces.top, wg.spaces.bottom
    if left_width > 0:
        yield Rect(wg.left - left_width, wg.top - top_height, wg.left, wg.bottom + bottom_height)
    if top_height > 0:
        yield Rect(wg.left, wg.top - top_height, wg.right + right_width, wg.top)
    if right_width > 0:
        yield Rect(wg.right, wg.top, wg.right + right_width, wg.bottom + bottom_height)
    if bottom_height > 0:
        yield Rect(wg.left, wg.bottom, wg.right, wg.bottom + bottom_height)


def window_geometry(xstart: int, xnum: int, ystart: int, ynum: int, left: int, top: int, right: int, bottom: int) -> WindowGeometry:
    return WindowGeometry(
        left=xstart, top=ystart, xnum=max(0, xnum), ynum=max(0, ynum),
        right=xstart + lgd.cell_width * xnum, bottom=ystart + lgd.cell_height * ynum,
        spaces=Edges(left, top, right, bottom)
    )


def window_geometry_from_layouts(x: LayoutData, y: LayoutData) -> WindowGeometry:
    return window_geometry(x.content_pos, x.cells_per_window, y.content_pos, y.cells_per_window, x.space_before, y.space_before, x.space_after, y.space_after)


def layout_single_window(
    xdecoration_pairs: DecorationPairs,
    ydecoration_pairs: DecorationPairs,
    xalignment: int = 0,
    yalignment: int = 0,
) -> WindowGeometry:
    x = next(layout_dimension(lgd.central.left, lgd.central.width, lgd.cell_width, xdecoration_pairs, alignment=xalignment))
    y = next(layout_dimension(lgd.central.top, lgd.central.height, lgd.cell_height, ydecoration_pairs, alignment=yalignment))
    return window_geometry_from_layouts(x, y)


def safe_increment_bias(old_val: float, increment: float = 0) -> float:
    return max(0.1, min(old_val + increment, 0.9))


def normalize_biases(biases: list[float]) -> list[float]:
    s = sum(biases)
    if s == 1.0:
        return biases
    return [x/s for x in biases]


def distribute_indexed_bias(base_bias: Sequence[float], index_bias_map: dict[int, float]) -> Sequence[float]:
    if not index_bias_map:
        return base_bias
    ans = list(base_bias)
    limit = len(ans)
    for row, increment in index_bias_map.items():
        if row >= limit or not increment:
            continue
        other_increment = -increment / (limit - 1)
        ans = [safe_increment_bias(b, increment if i == row else other_increment) for i, b in enumerate(ans)]
    return normalize_biases(ans)


def create_window_id_map_for_unserialize(all_windows: WindowList) -> dict[int, int]:
    window_id_map = {}
    for w in all_windows:
        if w.serialized_id:
            window_id_map[w.serialized_id] = w.id
    return window_id_map


class Layout:

    name: str = ''
    needs_window_borders = True
    must_draw_borders = False  # can be overridden to customize behavior from kittens
    layout_opts = LayoutOpts({})
    only_active_window_visible = False

    def __init__(self, os_window_id: int, tab_id: int, layout_opts: str = '') -> None:
        self.set_owner(os_window_id, tab_id)
        # A set of rectangles corresponding to the blank spaces at the edges of
        # this layout, i.e. spaces that are not covered by any window
        self.blank_rects: list[Rect] = []
        self.layout_opts = self.parse_layout_opts(layout_opts)
        assert self.name is not None
        self.full_name = f'{self.name}:{layout_opts}' if layout_opts else self.name
        self.remove_all_biases()

    def set_owner(self, os_window_id: int, tab_id: int) -> None:
        # Useful when moving a layout from one tab to another typically a detached tab being re-attached
        self.os_window_id = os_window_id
        self.tab_id = tab_id
        self.set_active_window_in_os_window = partial(set_active_window, os_window_id, tab_id)

    def bias_increment_for_cell(self, all_windows: WindowList, is_horizontal: bool) -> float:
        self._set_dimensions(all_windows)
        return self.calculate_bias_increment_for_a_single_cell(all_windows, is_horizontal)

    def calculate_bias_increment_for_a_single_cell(self, all_windows: WindowList, is_horizontal: bool) -> float:
        if is_horizontal:
            return (lgd.cell_width + 1) / lgd.central.width
        return (lgd.cell_height + 1) / lgd.central.height

    def apply_bias(self, window_id: int, increment: float, all_windows: WindowList, is_horizontal: bool = True) -> bool:
        return False

    def remove_all_biases(self) -> bool:
        return False

    def modify_size_of_window(self, all_windows: WindowList, window_id: int, increment: float, is_horizontal: bool = True) -> bool:
        idx = all_windows.group_idx_for_window(window_id)
        if idx is None or not increment:
            return False
        return self.apply_bias(idx, increment, all_windows, is_horizontal)

    def parse_layout_opts(self, layout_opts: str | None = None) -> LayoutOpts:
        data: dict[str, str] = {}
        if layout_opts:
            for x in layout_opts.split(';'):
                k, v = x.partition('=')[::2]
                if k and v:
                    data[k] = v
        return type(self.layout_opts)(data)

    def nth_window(self, all_windows: WindowList, num: int) -> WindowType | None:
        return all_windows.active_window_in_nth_group(num, clamp=True)

    def activate_nth_window(self, all_windows: WindowList, num: int) -> None:
        all_windows.set_active_group_idx(num)

    def next_window(self, all_windows: WindowList, delta: int = 1) -> None:
        all_windows.activate_next_window_group(delta)

    def neighbors(self, all_windows: WindowList) -> NeighborsMap:
        w = all_windows.active_window
        assert w is not None
        return self.neighbors_for_window(w, all_windows)

    def move_window(self, all_windows: WindowList, delta: int = 1) -> bool:
        if all_windows.num_groups < 2 or not delta:
            return False

        return all_windows.move_window_group(by=delta)

    def move_window_to_group(self, all_windows: WindowList, group: int) -> bool:
        return all_windows.move_window_group(to_group=group)

    def add_window(
        self, all_windows: WindowList, window: WindowType, location: str | None = None,
        overlay_for: int | None = None, put_overlay_behind: bool = False, bias: float | None = None,
        next_to: WindowType | None = None,
    ) -> WindowType | None:
        if overlay_for is not None:
            underlay = all_windows.id_map.get(overlay_for)
            if underlay is not None:
                window.margin, window.padding = underlay.margin.copy(), underlay.padding.copy()
                all_windows.add_window(window, group_of=overlay_for, head_of_group=put_overlay_behind)
                return underlay
        if location == 'neighbor':
            location = 'after'
        self.add_non_overlay_window(all_windows, window, location, bias, next_to)
        return None

    def add_non_overlay_window(
        self, all_windows: WindowList, window: WindowType, location: str | None, bias: float | None = None, next_to: WindowType | None = None
    ) -> None:
        before = False
        next_to = next_to or all_windows.active_window
        if location is not None:
            if location in ('after', 'vsplit', 'hsplit'):
                pass
            elif location == 'before':
                before = True
            elif location == 'first':
                before = True
                next_to = None
            elif location == 'last':
                next_to = None
        all_windows.add_window(window, next_to=next_to, before=before)
        if bias is not None:
            idx = all_windows.group_idx_for_window(window)
            if idx is not None:
                self._set_dimensions(all_windows)
                self._bias_slot(all_windows, idx, bias)

    def _bias_slot(self, all_windows: WindowList, idx: int, bias: float) -> bool:
        fractional_bias = max(10, min(abs(bias), 90)) / 100
        h, v = self.calculate_bias_increment_for_a_single_cell(all_windows, True), self.calculate_bias_increment_for_a_single_cell(all_windows, False)
        nh, nv = lgd.central.width / lgd.cell_width, lgd.central.height / lgd.cell_height
        f = max(-90, min(bias, 90)) / 100.
        return self.bias_slot(all_windows, idx, fractional_bias, h * nh *f, v * nv * f)

    def bias_slot(self, all_windows: WindowList, idx: int, fractional_bias: float, cell_increment_bias_h: float, cell_increment_bias_v: float) -> bool:
        return False

    def update_visibility(self, all_windows: WindowList) -> None:
        active_window = all_windows.active_window
        for window, is_group_leader in all_windows.iter_windows_with_visibility():
            is_visible = window is active_window or (is_group_leader and not self.only_active_window_visible)
            window.set_visible_in_layout(is_visible)

    def _set_dimensions(self, all_windows: WindowList) -> None:
        lgd.central, tab_bar, vw, vh, lgd.cell_width, lgd.cell_height = viewport_for_window(self.os_window_id)
        # Update lgd.draw_minimal_borders based on the current number of visible windows
        # and the draw_window_borders_for_single_window option
        opts = get_options()
        lgd.draw_minimal_borders = effective_draw_minimal_borders(opts, all_windows.has_more_than_one_visible_group)

    def __call__(self, all_windows: WindowList) -> None:
        self._set_dimensions(all_windows)
        self.update_visibility(all_windows)
        self.blank_rects = []
        self.do_layout(all_windows)

    def layout_single_window_group(self, wg: WindowGroup, add_blank_rects: bool = True) -> None:
        bw = 1 if self.must_draw_borders else 0
        xdecoration_pairs = ((
            wg.decoration('left', border_mult=bw, is_single_window=True),
            wg.decoration('right', border_mult=bw, is_single_window=True),
        ),)
        ydecoration_pairs = ((
            wg.decoration('top', border_mult=bw, is_single_window=True),
            wg.decoration('bottom', border_mult=bw, is_single_window=True),
        ),)
        geom = layout_single_window(xdecoration_pairs, ydecoration_pairs, xalignment=lgd.alignment_x, yalignment=lgd.alignment_y)
        wg.set_geometry(geom)
        if add_blank_rects:
            self.blank_rects.extend(blank_rects_for_window(geom))

    def xlayout(
        self,
        groups: Iterator[WindowGroup],
        bias: None | Sequence[float] | dict[int, float] = None,
        start: int | None = None,
        size: int | None = None,
        offset: int = 0,
        border_mult: int = 1
    ) -> LayoutDimension:
        decoration_pairs = tuple(
            (g.decoration('left', border_mult=border_mult), g.decoration('right', border_mult=border_mult)) for i, g in
            enumerate(groups) if i >= offset
        )
        if start is None:
            start = lgd.central.left
        if size is None:
            size = lgd.central.width
        return layout_dimension(start, size, lgd.cell_width, decoration_pairs, bias=bias, alignment=lgd.alignment_x)

    def ylayout(
        self,
        groups: Iterator[WindowGroup],
        bias: None | Sequence[float] | dict[int, float] = None,
        start: int | None = None,
        size: int | None = None,
        offset: int = 0,
        border_mult: int = 1
    ) -> LayoutDimension:
        decoration_pairs = tuple(
            (g.decoration('top', border_mult=border_mult), g.decoration('bottom', border_mult=border_mult)) for i, g in
            enumerate(groups) if i >= offset
        )
        if start is None:
            start = lgd.central.top
        if size is None:
            size = lgd.central.height
        return layout_dimension(start, size, lgd.cell_height, decoration_pairs, bias=bias, alignment=lgd.alignment_y)

    def set_window_group_geometry(self, wg: WindowGroup, xl: LayoutData, yl: LayoutData) -> WindowGeometry:
        geom = window_geometry_from_layouts(xl, yl)
        wg.set_geometry(geom)
        self.blank_rects.extend(blank_rects_for_window(geom))
        return geom

    def do_layout(self, windows: WindowList) -> None:
        raise NotImplementedError()

    def neighbors_for_window(self, window: WindowType, windows: WindowList) -> NeighborsMap:
        return {}

    def compute_needs_borders_map(self, all_windows: WindowList) -> dict[int, bool]:
        return all_windows.compute_needs_borders_map(lgd.draw_active_borders)

    def get_minimal_borders(self, windows: WindowList) -> Generator[BorderLine, None, None]:
        self._set_dimensions(windows)
        yield from self.minimal_borders(windows)

    def minimal_borders(self, windows: WindowList) -> Generator[BorderLine, None, None]:
        return
        yield BorderLine()  # type: ignore

    def layout_action(self, action_name: str, args: Sequence[str], all_windows: WindowList) -> bool | None:
        pass

    def layout_state(self) -> dict[str, Any]:
        return {}

    def set_layout_state(self, layout_state: dict[str, Any], map_group_id: WindowMapper) -> bool:
        return True

    def serialize(self, all_windows: WindowList) -> dict[str, Any]:
        ans = self.layout_state()
        ans['opts'] = self.layout_opts.serialized()
        ans['class'] = self.__class__.__name__
        ans['all_windows'] = all_windows.serialize_layout_state()
        return ans

    def unserialize(
        self, s: dict[str, Any], all_windows: WindowList,
        window_id_mapper: Callable[[WindowList], dict[int, int]] = create_window_id_map_for_unserialize,
    ) -> bool:
        if s.get('class') != self.__class__.__name__:
            return False
        window_id_map = create_window_id_map_for_unserialize(all_windows)
        m = all_windows.unserialize_layout_state(s['all_windows'], window_id_map)
        if m is None:
            return False
        return self.set_layout_state(s, m.get)



# generated by gen-config.py DO NOT edit

# isort: skip_file
import typing
import collections.abc  # noqa: F401, RUF100
from kitty.conf.utils import (
    merge_dicts, positive_float, positive_int, python_string, to_bool, to_cmdline, to_color,
    to_color_or_none, unit_float
)
from kitty.options.utils import (
    action_alias, active_tab_title_template, allow_hyperlinks, bell_on_tab, box_drawing_scale,
    clear_all_mouse_actions, clear_all_shortcuts, clipboard_control, clone_source_strategies,
    config_or_absolute_path, confirm_close, copy_on_select, cursor_blink_interval, cursor_text_color,
    cursor_trail_decay, deprecated_adjust_line_height, deprecated_hide_window_decorations_aliases,
    deprecated_macos_show_window_title_in_menubar_alias, deprecated_scrollback_indicator_opacity,
    deprecated_send_text, disable_ligatures, edge_width, env, filter_notification, font_features,
    hide_window_decorations, macos_option_as_alt, macos_titlebar_color, menu_map, modify_font,
    mouse_hide_wait, narrow_symbols, notify_on_cmd_finish, optional_edge_width, parse_font_spec,
    parse_map, parse_mouse_map, paste_actions, pointer_shape_when_dragging, remote_control_password,
    resize_debounce_time, scrollback_lines, scrollback_pager_history_size, scrollbar_color,
    shell_integration, store_multiple, symbol_map, tab_activity_symbol, tab_bar_edge,
    tab_bar_margin_height, tab_bar_min_tabs, tab_fade, tab_font_style, tab_separator,
    tab_title_template, text_fg_override_threshold, titlebar_color, to_cursor_shape,
    to_cursor_unfocused_shape, to_font_size, to_layout_names, to_modifiers,
    transparent_background_colors, underline_exclusion, url_prefixes, url_style, visual_bell_duration,
    visual_window_select_characters, window_border_width, window_logo_scale, window_size
)


class Parser:

    def action_alias(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in action_alias(val):
            ans["action_alias"][k] = v

    def active_border_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['active_border_color'] = to_color_or_none(val)

    def active_tab_background(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['active_tab_background'] = to_color(val)

    def active_tab_font_style(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['active_tab_font_style'] = tab_font_style(val)

    def active_tab_foreground(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['active_tab_foreground'] = to_color(val)

    def active_tab_title_template(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['active_tab_title_template'] = active_tab_title_template(val)

    def allow_cloning(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_allow_cloning:
            raise ValueError(f"The value {val} is not a valid choice for allow_cloning")
        ans["allow_cloning"] = val

    choices_for_allow_cloning = frozenset(('yes', 'y', 'true', 'no', 'n', 'false', 'ask'))

    def allow_hyperlinks(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['allow_hyperlinks'] = allow_hyperlinks(val)

    def allow_remote_control(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_allow_remote_control:
            raise ValueError(f"The value {val} is not a valid choice for allow_remote_control")
        ans["allow_remote_control"] = val

    choices_for_allow_remote_control = frozenset(('password', 'socket-only', 'socket', 'no', 'n', 'false', 'yes', 'y', 'true'))

    def background(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['background'] = to_color(val)

    def background_blur(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['background_blur'] = int(val)

    def background_image(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['background_image'] = config_or_absolute_path(val)

    def background_image_layout(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_background_image_layout:
            raise ValueError(f"The value {val} is not a valid choice for background_image_layout")
        ans["background_image_layout"] = val

    choices_for_background_image_layout = frozenset(('mirror-tiled', 'scaled', 'tiled', 'clamped', 'centered', 'cscaled'))

    def background_image_linear(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['background_image_linear'] = to_bool(val)

    def background_opacity(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['background_opacity'] = unit_float(val)

    def background_tint(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['background_tint'] = unit_float(val)

    def background_tint_gaps(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['background_tint_gaps'] = unit_float(val)

    def bell_border_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['bell_border_color'] = to_color(val)

    def bell_on_tab(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['bell_on_tab'] = bell_on_tab(val)

    def bell_path(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['bell_path'] = config_or_absolute_path(val)

    def bold_font(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['bold_font'] = parse_font_spec(val)

    def bold_italic_font(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['bold_italic_font'] = parse_font_spec(val)

    def box_drawing_scale(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['box_drawing_scale'] = box_drawing_scale(val)

    def clear_all_mouse_actions(self, val: str, ans: dict[str, typing.Any]) -> None:
        clear_all_mouse_actions(val, ans)

    def clear_all_shortcuts(self, val: str, ans: dict[str, typing.Any]) -> None:
        clear_all_shortcuts(val, ans)

    def clear_selection_on_clipboard_loss(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['clear_selection_on_clipboard_loss'] = to_bool(val)

    def click_interval(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['click_interval'] = float(val)

    def clipboard_control(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['clipboard_control'] = clipboard_control(val)

    def clipboard_max_size(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['clipboard_max_size'] = positive_float(val)

    def clone_source_strategies(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['clone_source_strategies'] = clone_source_strategies(val)

    def close_on_child_death(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['close_on_child_death'] = to_bool(val)

    def color0(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color0'] = to_color(val)

    def color1(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color1'] = to_color(val)

    def color2(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color2'] = to_color(val)

    def color3(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color3'] = to_color(val)

    def color4(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color4'] = to_color(val)

    def color5(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color5'] = to_color(val)

    def color6(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color6'] = to_color(val)

    def color7(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color7'] = to_color(val)

    def color8(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color8'] = to_color(val)

    def color9(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color9'] = to_color(val)

    def color10(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color10'] = to_color(val)

    def color11(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color11'] = to_color(val)

    def color12(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color12'] = to_color(val)

    def color13(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color13'] = to_color(val)

    def color14(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color14'] = to_color(val)

    def color15(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color15'] = to_color(val)

    def color16(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color16'] = to_color(val)

    def color17(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color17'] = to_color(val)

    def color18(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color18'] = to_color(val)

    def color19(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color19'] = to_color(val)

    def color20(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color20'] = to_color(val)

    def color21(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color21'] = to_color(val)

    def color22(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color22'] = to_color(val)

    def color23(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color23'] = to_color(val)

    def color24(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color24'] = to_color(val)

    def color25(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color25'] = to_color(val)

    def color26(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color26'] = to_color(val)

    def color27(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color27'] = to_color(val)

    def color28(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color28'] = to_color(val)

    def color29(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color29'] = to_color(val)

    def color30(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color30'] = to_color(val)

    def color31(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color31'] = to_color(val)

    def color32(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color32'] = to_color(val)

    def color33(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color33'] = to_color(val)

    def color34(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color34'] = to_color(val)

    def color35(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color35'] = to_color(val)

    def color36(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color36'] = to_color(val)

    def color37(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color37'] = to_color(val)

    def color38(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color38'] = to_color(val)

    def color39(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color39'] = to_color(val)

    def color40(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color40'] = to_color(val)

    def color41(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color41'] = to_color(val)

    def color42(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color42'] = to_color(val)

    def color43(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color43'] = to_color(val)

    def color44(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color44'] = to_color(val)

    def color45(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color45'] = to_color(val)

    def color46(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color46'] = to_color(val)

    def color47(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color47'] = to_color(val)

    def color48(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color48'] = to_color(val)

    def color49(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color49'] = to_color(val)

    def color50(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color50'] = to_color(val)

    def color51(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color51'] = to_color(val)

    def color52(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color52'] = to_color(val)

    def color53(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color53'] = to_color(val)

    def color54(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color54'] = to_color(val)

    def color55(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color55'] = to_color(val)

    def color56(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color56'] = to_color(val)

    def color57(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color57'] = to_color(val)

    def color58(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color58'] = to_color(val)

    def color59(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color59'] = to_color(val)

    def color60(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color60'] = to_color(val)

    def color61(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color61'] = to_color(val)

    def color62(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color62'] = to_color(val)

    def color63(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color63'] = to_color(val)

    def color64(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color64'] = to_color(val)

    def color65(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color65'] = to_color(val)

    def color66(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color66'] = to_color(val)

    def color67(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color67'] = to_color(val)

    def color68(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color68'] = to_color(val)

    def color69(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color69'] = to_color(val)

    def color70(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color70'] = to_color(val)

    def color71(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color71'] = to_color(val)

    def color72(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color72'] = to_color(val)

    def color73(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color73'] = to_color(val)

    def color74(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color74'] = to_color(val)

    def color75(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color75'] = to_color(val)

    def color76(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color76'] = to_color(val)

    def color77(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color77'] = to_color(val)

    def color78(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color78'] = to_color(val)

    def color79(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color79'] = to_color(val)

    def color80(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color80'] = to_color(val)

    def color81(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color81'] = to_color(val)

    def color82(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color82'] = to_color(val)

    def color83(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color83'] = to_color(val)

    def color84(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color84'] = to_color(val)

    def color85(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color85'] = to_color(val)

    def color86(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color86'] = to_color(val)

    def color87(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color87'] = to_color(val)

    def color88(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color88'] = to_color(val)

    def color89(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color89'] = to_color(val)

    def color90(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color90'] = to_color(val)

    def color91(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color91'] = to_color(val)

    def color92(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color92'] = to_color(val)

    def color93(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color93'] = to_color(val)

    def color94(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color94'] = to_color(val)

    def color95(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color95'] = to_color(val)

    def color96(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color96'] = to_color(val)

    def color97(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color97'] = to_color(val)

    def color98(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color98'] = to_color(val)

    def color99(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color99'] = to_color(val)

    def color100(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color100'] = to_color(val)

    def color101(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color101'] = to_color(val)

    def color102(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color102'] = to_color(val)

    def color103(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color103'] = to_color(val)

    def color104(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color104'] = to_color(val)

    def color105(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color105'] = to_color(val)

    def color106(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color106'] = to_color(val)

    def color107(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color107'] = to_color(val)

    def color108(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color108'] = to_color(val)

    def color109(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color109'] = to_color(val)

    def color110(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color110'] = to_color(val)

    def color111(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color111'] = to_color(val)

    def color112(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color112'] = to_color(val)

    def color113(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color113'] = to_color(val)

    def color114(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color114'] = to_color(val)

    def color115(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color115'] = to_color(val)

    def color116(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color116'] = to_color(val)

    def color117(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color117'] = to_color(val)

    def color118(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color118'] = to_color(val)

    def color119(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color119'] = to_color(val)

    def color120(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color120'] = to_color(val)

    def color121(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color121'] = to_color(val)

    def color122(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color122'] = to_color(val)

    def color123(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color123'] = to_color(val)

    def color124(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color124'] = to_color(val)

    def color125(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color125'] = to_color(val)

    def color126(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color126'] = to_color(val)

    def color127(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color127'] = to_color(val)

    def color128(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color128'] = to_color(val)

    def color129(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color129'] = to_color(val)

    def color130(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color130'] = to_color(val)

    def color131(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color131'] = to_color(val)

    def color132(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color132'] = to_color(val)

    def color133(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color133'] = to_color(val)

    def color134(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color134'] = to_color(val)

    def color135(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color135'] = to_color(val)

    def color136(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color136'] = to_color(val)

    def color137(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color137'] = to_color(val)

    def color138(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color138'] = to_color(val)

    def color139(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color139'] = to_color(val)

    def color140(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color140'] = to_color(val)

    def color141(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color141'] = to_color(val)

    def color142(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color142'] = to_color(val)

    def color143(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color143'] = to_color(val)

    def color144(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color144'] = to_color(val)

    def color145(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color145'] = to_color(val)

    def color146(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color146'] = to_color(val)

    def color147(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color147'] = to_color(val)

    def color148(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color148'] = to_color(val)

    def color149(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color149'] = to_color(val)

    def color150(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color150'] = to_color(val)

    def color151(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color151'] = to_color(val)

    def color152(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color152'] = to_color(val)

    def color153(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color153'] = to_color(val)

    def color154(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color154'] = to_color(val)

    def color155(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color155'] = to_color(val)

    def color156(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color156'] = to_color(val)

    def color157(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color157'] = to_color(val)

    def color158(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color158'] = to_color(val)

    def color159(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color159'] = to_color(val)

    def color160(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color160'] = to_color(val)

    def color161(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color161'] = to_color(val)

    def color162(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color162'] = to_color(val)

    def color163(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color163'] = to_color(val)

    def color164(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color164'] = to_color(val)

    def color165(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color165'] = to_color(val)

    def color166(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color166'] = to_color(val)

    def color167(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color167'] = to_color(val)

    def color168(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color168'] = to_color(val)

    def color169(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color169'] = to_color(val)

    def color170(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color170'] = to_color(val)

    def color171(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color171'] = to_color(val)

    def color172(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color172'] = to_color(val)

    def color173(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color173'] = to_color(val)

    def color174(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color174'] = to_color(val)

    def color175(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color175'] = to_color(val)

    def color176(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color176'] = to_color(val)

    def color177(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color177'] = to_color(val)

    def color178(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color178'] = to_color(val)

    def color179(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color179'] = to_color(val)

    def color180(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color180'] = to_color(val)

    def color181(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color181'] = to_color(val)

    def color182(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color182'] = to_color(val)

    def color183(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color183'] = to_color(val)

    def color184(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color184'] = to_color(val)

    def color185(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color185'] = to_color(val)

    def color186(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color186'] = to_color(val)

    def color187(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color187'] = to_color(val)

    def color188(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color188'] = to_color(val)

    def color189(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color189'] = to_color(val)

    def color190(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color190'] = to_color(val)

    def color191(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color191'] = to_color(val)

    def color192(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color192'] = to_color(val)

    def color193(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color193'] = to_color(val)

    def color194(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color194'] = to_color(val)

    def color195(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color195'] = to_color(val)

    def color196(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color196'] = to_color(val)

    def color197(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color197'] = to_color(val)

    def color198(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color198'] = to_color(val)

    def color199(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color199'] = to_color(val)

    def color200(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color200'] = to_color(val)

    def color201(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color201'] = to_color(val)

    def color202(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color202'] = to_color(val)

    def color203(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color203'] = to_color(val)

    def color204(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color204'] = to_color(val)

    def color205(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color205'] = to_color(val)

    def color206(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color206'] = to_color(val)

    def color207(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color207'] = to_color(val)

    def color208(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color208'] = to_color(val)

    def color209(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color209'] = to_color(val)

    def color210(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color210'] = to_color(val)

    def color211(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color211'] = to_color(val)

    def color212(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color212'] = to_color(val)

    def color213(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color213'] = to_color(val)

    def color214(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color214'] = to_color(val)

    def color215(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color215'] = to_color(val)

    def color216(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color216'] = to_color(val)

    def color217(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color217'] = to_color(val)

    def color218(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color218'] = to_color(val)

    def color219(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color219'] = to_color(val)

    def color220(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color220'] = to_color(val)

    def color221(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color221'] = to_color(val)

    def color222(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color222'] = to_color(val)

    def color223(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color223'] = to_color(val)

    def color224(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color224'] = to_color(val)

    def color225(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color225'] = to_color(val)

    def color226(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color226'] = to_color(val)

    def color227(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color227'] = to_color(val)

    def color228(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color228'] = to_color(val)

    def color229(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color229'] = to_color(val)

    def color230(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color230'] = to_color(val)

    def color231(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color231'] = to_color(val)

    def color232(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color232'] = to_color(val)

    def color233(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color233'] = to_color(val)

    def color234(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color234'] = to_color(val)

    def color235(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color235'] = to_color(val)

    def color236(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color236'] = to_color(val)

    def color237(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color237'] = to_color(val)

    def color238(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color238'] = to_color(val)

    def color239(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color239'] = to_color(val)

    def color240(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color240'] = to_color(val)

    def color241(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color241'] = to_color(val)

    def color242(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color242'] = to_color(val)

    def color243(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color243'] = to_color(val)

    def color244(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color244'] = to_color(val)

    def color245(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color245'] = to_color(val)

    def color246(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color246'] = to_color(val)

    def color247(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color247'] = to_color(val)

    def color248(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color248'] = to_color(val)

    def color249(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color249'] = to_color(val)

    def color250(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color250'] = to_color(val)

    def color251(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color251'] = to_color(val)

    def color252(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color252'] = to_color(val)

    def color253(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color253'] = to_color(val)

    def color254(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color254'] = to_color(val)

    def color255(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['color255'] = to_color(val)

    def command_on_bell(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['command_on_bell'] = to_cmdline(val)

    def confirm_os_window_close(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['confirm_os_window_close'] = confirm_close(val)

    def copy_on_select(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['copy_on_select'] = copy_on_select(val)

    def cursor(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor'] = to_color_or_none(val)

    def cursor_beam_thickness(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_beam_thickness'] = positive_float(val)

    def cursor_blink_interval(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_blink_interval'] = cursor_blink_interval(val)

    def cursor_shape(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_shape'] = to_cursor_shape(val)

    def cursor_shape_unfocused(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_shape_unfocused'] = to_cursor_unfocused_shape(val)

    def cursor_stop_blinking_after(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_stop_blinking_after'] = positive_float(val)

    def cursor_text_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_text_color'] = cursor_text_color(val)

    def cursor_trail(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_trail'] = positive_int(val)

    def cursor_trail_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_trail_color'] = to_color_or_none(val)

    def cursor_trail_decay(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_trail_decay'] = cursor_trail_decay(val)

    def cursor_trail_start_threshold(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_trail_start_threshold'] = positive_int(val)

    def cursor_underline_thickness(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['cursor_underline_thickness'] = positive_float(val)

    def default_pointer_shape(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_default_pointer_shape:
            raise ValueError(f"The value {val} is not a valid choice for default_pointer_shape")
        ans["default_pointer_shape"] = val

    choices_for_default_pointer_shape = frozenset(('arrow', 'beam', 'text', 'pointer', 'hand', 'help', 'wait', 'progress', 'crosshair', 'cell', 'vertical-text', 'move', 'e-resize', 'ne-resize', 'nw-resize', 'n-resize', 'se-resize', 'sw-resize', 's-resize', 'w-resize', 'ew-resize', 'ns-resize', 'nesw-resize', 'nwse-resize', 'zoom-in', 'zoom-out', 'alias', 'copy', 'not-allowed', 'no-drop', 'grab', 'grabbing'))

    def detect_urls(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['detect_urls'] = to_bool(val)

    def dim_opacity(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['dim_opacity'] = unit_float(val)

    def disable_ligatures(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['disable_ligatures'] = disable_ligatures(val)

    def draw_minimal_borders(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['draw_minimal_borders'] = to_bool(val)

    def draw_window_borders_for_single_window(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['draw_window_borders_for_single_window'] = to_bool(val)

    def dynamic_background_opacity(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['dynamic_background_opacity'] = to_bool(val)

    def editor(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['editor'] = str(val)

    def enable_audio_bell(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['enable_audio_bell'] = to_bool(val)

    def enabled_layouts(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['enabled_layouts'] = to_layout_names(val)

    def env(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in env(val, ans["env"]):
            ans["env"][k] = v

    def exe_search_path(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in store_multiple(val, ans["exe_search_path"]):
            ans["exe_search_path"][k] = v

    def file_transfer_confirmation_bypass(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['file_transfer_confirmation_bypass'] = str(val)

    def filter_notification(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in filter_notification(val, ans["filter_notification"]):
            ans["filter_notification"][k] = v

    def focus_follows_mouse(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['focus_follows_mouse'] = to_bool(val)

    def font_family(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['font_family'] = parse_font_spec(val)

    def font_features(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in font_features(val):
            ans["font_features"][k] = v

    def font_size(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['font_size'] = to_font_size(val)

    def force_ltr(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['force_ltr'] = to_bool(val)

    def foreground(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['foreground'] = to_color(val)

    def forward_stdio(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['forward_stdio'] = to_bool(val)

    def hide_window_decorations(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['hide_window_decorations'] = hide_window_decorations(val)

    def inactive_border_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['inactive_border_color'] = to_color(val)

    def inactive_tab_background(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['inactive_tab_background'] = to_color(val)

    def inactive_tab_font_style(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['inactive_tab_font_style'] = tab_font_style(val)

    def inactive_tab_foreground(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['inactive_tab_foreground'] = to_color(val)

    def inactive_text_alpha(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['inactive_text_alpha'] = unit_float(val)

    def initial_window_height(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['initial_window_height'] = window_size(val)

    def initial_window_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['initial_window_width'] = window_size(val)

    def input_delay(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['input_delay'] = positive_int(val)

    def italic_font(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['italic_font'] = parse_font_spec(val)

    def kitten_alias(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in action_alias(val):
            ans["kitten_alias"][k] = v

    def kitty_mod(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['kitty_mod'] = to_modifiers(val)

    def linux_bell_theme(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['linux_bell_theme'] = str(val)

    def linux_display_server(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_linux_display_server:
            raise ValueError(f"The value {val} is not a valid choice for linux_display_server")
        ans["linux_display_server"] = val

    choices_for_linux_display_server = frozenset(('auto', 'wayland', 'x11'))

    def listen_on(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['listen_on'] = str(val)

    def macos_colorspace(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_macos_colorspace:
            raise ValueError(f"The value {val} is not a valid choice for macos_colorspace")
        ans["macos_colorspace"] = val

    choices_for_macos_colorspace = frozenset(('srgb', 'default', 'displayp3'))

    def macos_custom_beam_cursor(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['macos_custom_beam_cursor'] = to_bool(val)

    def macos_hide_from_tasks(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['macos_hide_from_tasks'] = to_bool(val)

    def macos_menubar_title_max_length(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['macos_menubar_title_max_length'] = positive_int(val)

    def macos_option_as_alt(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['macos_option_as_alt'] = macos_option_as_alt(val)

    def macos_quit_when_last_window_closed(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['macos_quit_when_last_window_closed'] = to_bool(val)

    def macos_show_window_title_in(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_macos_show_window_title_in:
            raise ValueError(f"The value {val} is not a valid choice for macos_show_window_title_in")
        ans["macos_show_window_title_in"] = val

    choices_for_macos_show_window_title_in = frozenset(('all', 'menubar', 'none', 'window'))

    def macos_thicken_font(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['macos_thicken_font'] = positive_float(val)

    def macos_titlebar_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['macos_titlebar_color'] = macos_titlebar_color(val)

    def macos_traditional_fullscreen(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['macos_traditional_fullscreen'] = to_bool(val)

    def macos_window_resizable(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['macos_window_resizable'] = to_bool(val)

    def mark1_background(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['mark1_background'] = to_color(val)

    def mark1_foreground(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['mark1_foreground'] = to_color(val)

    def mark2_background(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['mark2_background'] = to_color(val)

    def mark2_foreground(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['mark2_foreground'] = to_color(val)

    def mark3_background(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['mark3_background'] = to_color(val)

    def mark3_foreground(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['mark3_foreground'] = to_color(val)

    def menu_map(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in menu_map(val, ans["menu_map"]):
            ans["menu_map"][k] = v

    def modify_font(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in modify_font(val):
            ans["modify_font"][k] = v

    def mouse_hide_wait(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['mouse_hide_wait'] = mouse_hide_wait(val)

    def narrow_symbols(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in narrow_symbols(val):
            ans["narrow_symbols"][k] = v

    def notify_on_cmd_finish(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['notify_on_cmd_finish'] = notify_on_cmd_finish(val)

    def open_url_with(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['open_url_with'] = to_cmdline(val)

    def paste_actions(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['paste_actions'] = paste_actions(val)

    def placement_strategy(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_placement_strategy:
            raise ValueError(f"The value {val} is not a valid choice for placement_strategy")
        ans["placement_strategy"] = val

    choices_for_placement_strategy = frozenset(('top-left', 'top', 'top-right', 'left', 'center', 'right', 'bottom-left', 'bottom', 'bottom-right'))

    def pointer_shape_when_dragging(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['pointer_shape_when_dragging'] = pointer_shape_when_dragging(val)

    def pointer_shape_when_grabbed(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_pointer_shape_when_grabbed:
            raise ValueError(f"The value {val} is not a valid choice for pointer_shape_when_grabbed")
        ans["pointer_shape_when_grabbed"] = val

    choices_for_pointer_shape_when_grabbed = choices_for_default_pointer_shape

    def remember_window_position(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['remember_window_position'] = to_bool(val)

    def remember_window_size(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['remember_window_size'] = to_bool(val)

    def remote_control_password(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in remote_control_password(val, ans["remote_control_password"]):
            ans["remote_control_password"][k] = v

    def repaint_delay(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['repaint_delay'] = positive_int(val)

    def resize_debounce_time(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['resize_debounce_time'] = resize_debounce_time(val)

    def resize_in_steps(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['resize_in_steps'] = to_bool(val)

    def scrollback_fill_enlarged_window(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollback_fill_enlarged_window'] = to_bool(val)

    def scrollback_lines(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollback_lines'] = scrollback_lines(val)

    def scrollback_pager(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollback_pager'] = to_cmdline(val)

    def scrollback_pager_history_size(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollback_pager_history_size'] = scrollback_pager_history_size(val)

    def scrollbar(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_scrollbar:
            raise ValueError(f"The value {val} is not a valid choice for scrollbar")
        ans["scrollbar"] = val

    choices_for_scrollbar = frozenset(('scrolled', 'always', 'never', 'hovered', 'scrolled-and-hovered'))

    def scrollbar_gap(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_gap'] = positive_float(val)

    def scrollbar_handle_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_handle_color'] = scrollbar_color(val)

    def scrollbar_handle_opacity(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_handle_opacity'] = positive_float(val)

    def scrollbar_hitbox_expansion(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_hitbox_expansion'] = positive_float(val)

    def scrollbar_hover_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_hover_width'] = positive_float(val)

    def scrollbar_interactive(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_interactive'] = to_bool(val)

    def scrollbar_jump_on_click(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_jump_on_click'] = to_bool(val)

    def scrollbar_min_handle_height(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_min_handle_height'] = positive_float(val)

    def scrollbar_radius(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_radius'] = positive_float(val)

    def scrollbar_track_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_track_color'] = scrollbar_color(val)

    def scrollbar_track_hover_opacity(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_track_hover_opacity'] = positive_float(val)

    def scrollbar_track_opacity(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_track_opacity'] = positive_float(val)

    def scrollbar_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['scrollbar_width'] = positive_float(val)

    def select_by_word_characters(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['select_by_word_characters'] = str(val)

    def select_by_word_characters_forward(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['select_by_word_characters_forward'] = str(val)

    def selection_background(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['selection_background'] = to_color_or_none(val)

    def selection_foreground(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['selection_foreground'] = to_color_or_none(val)

    def shell(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['shell'] = str(val)

    def shell_integration(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['shell_integration'] = shell_integration(val)

    def show_hyperlink_targets(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['show_hyperlink_targets'] = to_bool(val)

    def single_window_margin_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['single_window_margin_width'] = optional_edge_width(val)

    def single_window_padding_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['single_window_padding_width'] = optional_edge_width(val)

    def startup_session(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['startup_session'] = config_or_absolute_path(val)

    def strip_trailing_spaces(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_strip_trailing_spaces:
            raise ValueError(f"The value {val} is not a valid choice for strip_trailing_spaces")
        ans["strip_trailing_spaces"] = val

    choices_for_strip_trailing_spaces = frozenset(('always', 'never', 'smart'))

    def symbol_map(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in symbol_map(val):
            ans["symbol_map"][k] = v

    def sync_to_monitor(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['sync_to_monitor'] = to_bool(val)

    def tab_activity_symbol(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_activity_symbol'] = tab_activity_symbol(val)

    def tab_bar_align(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_tab_bar_align:
            raise ValueError(f"The value {val} is not a valid choice for tab_bar_align")
        ans["tab_bar_align"] = val

    choices_for_tab_bar_align = frozenset(('left', 'center', 'right'))

    def tab_bar_background(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_bar_background'] = to_color_or_none(val)

    def tab_bar_edge(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_bar_edge'] = tab_bar_edge(val)

    def tab_bar_filter(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_bar_filter'] = str(val)

    def tab_bar_margin_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_bar_margin_color'] = to_color_or_none(val)

    def tab_bar_margin_height(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_bar_margin_height'] = tab_bar_margin_height(val)

    def tab_bar_margin_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_bar_margin_width'] = positive_float(val)

    def tab_bar_min_tabs(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_bar_min_tabs'] = tab_bar_min_tabs(val)

    def tab_bar_style(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_tab_bar_style:
            raise ValueError(f"The value {val} is not a valid choice for tab_bar_style")
        ans["tab_bar_style"] = val

    choices_for_tab_bar_style = frozenset(('fade', 'hidden', 'powerline', 'separator', 'slant', 'custom'))

    def tab_fade(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_fade'] = tab_fade(val)

    def tab_powerline_style(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_tab_powerline_style:
            raise ValueError(f"The value {val} is not a valid choice for tab_powerline_style")
        ans["tab_powerline_style"] = val

    choices_for_tab_powerline_style = frozenset(('angled', 'round', 'slanted'))

    def tab_separator(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_separator'] = tab_separator(val)

    def tab_switch_strategy(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_tab_switch_strategy:
            raise ValueError(f"The value {val} is not a valid choice for tab_switch_strategy")
        ans["tab_switch_strategy"] = val

    choices_for_tab_switch_strategy = frozenset(('last', 'left', 'previous', 'right'))

    def tab_title_max_length(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_title_max_length'] = positive_int(val)

    def tab_title_template(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['tab_title_template'] = tab_title_template(val)

    def term(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['term'] = str(val)

    def terminfo_type(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_terminfo_type:
            raise ValueError(f"The value {val} is not a valid choice for terminfo_type")
        ans["terminfo_type"] = val

    choices_for_terminfo_type = frozenset(('path', 'direct', 'none'))

    def text_composition_strategy(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['text_composition_strategy'] = str(val)

    def text_fg_override_threshold(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['text_fg_override_threshold'] = text_fg_override_threshold(val)

    def touch_scroll_multiplier(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['touch_scroll_multiplier'] = float(val)

    def transparent_background_colors(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['transparent_background_colors'] = transparent_background_colors(val)

    def undercurl_style(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_undercurl_style:
            raise ValueError(f"The value {val} is not a valid choice for undercurl_style")
        ans["undercurl_style"] = val

    choices_for_undercurl_style = frozenset(('thin-sparse', 'thin-dense', 'thick-sparse', 'thick-dense'))

    def underline_exclusion(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['underline_exclusion'] = underline_exclusion(val)

    def underline_hyperlinks(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_underline_hyperlinks:
            raise ValueError(f"The value {val} is not a valid choice for underline_hyperlinks")
        ans["underline_hyperlinks"] = val

    choices_for_underline_hyperlinks = frozenset(('hover', 'always', 'never'))

    def update_check_interval(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['update_check_interval'] = float(val)

    def url_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['url_color'] = to_color(val)

    def url_excluded_characters(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['url_excluded_characters'] = python_string(val)

    def url_prefixes(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['url_prefixes'] = url_prefixes(val)

    def url_style(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['url_style'] = url_style(val)

    def visual_bell_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['visual_bell_color'] = to_color_or_none(val)

    def visual_bell_duration(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['visual_bell_duration'] = visual_bell_duration(val)

    def visual_window_select_characters(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['visual_window_select_characters'] = visual_window_select_characters(val)

    def watcher(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k, v in store_multiple(val, ans["watcher"]):
            ans["watcher"][k] = v

    def wayland_enable_ime(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['wayland_enable_ime'] = to_bool(val)

    def wayland_titlebar_color(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['wayland_titlebar_color'] = titlebar_color(val)

    def wheel_scroll_min_lines(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['wheel_scroll_min_lines'] = int(val)

    def wheel_scroll_multiplier(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['wheel_scroll_multiplier'] = float(val)

    def window_alert_on_bell(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['window_alert_on_bell'] = to_bool(val)

    def window_border_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['window_border_width'] = window_border_width(val)

    def window_logo_alpha(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['window_logo_alpha'] = unit_float(val)

    def window_logo_path(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['window_logo_path'] = config_or_absolute_path(val)

    def window_logo_position(self, val: str, ans: dict[str, typing.Any]) -> None:
        val = val.lower()
        if val not in self.choices_for_window_logo_position:
            raise ValueError(f"The value {val} is not a valid choice for window_logo_position")
        ans["window_logo_position"] = val

    choices_for_window_logo_position = choices_for_placement_strategy

    def window_logo_scale(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['window_logo_scale'] = window_logo_scale(val)

    def window_margin_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['window_margin_width'] = edge_width(val)

    def window_padding_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['window_padding_width'] = edge_width(val)

    def window_resize_step_cells(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['window_resize_step_cells'] = positive_int(val)

    def window_resize_step_lines(self, val: str, ans: dict[str, typing.Any]) -> None:
        ans['window_resize_step_lines'] = positive_int(val)

    def x11_hide_window_decorations(self, val: str, ans: dict[str, typing.Any]) -> None:
        deprecated_hide_window_decorations_aliases('x11_hide_window_decorations', val, ans)

    def macos_hide_titlebar(self, val: str, ans: dict[str, typing.Any]) -> None:
        deprecated_hide_window_decorations_aliases('macos_hide_titlebar', val, ans)

    def macos_show_window_title_in_menubar(self, val: str, ans: dict[str, typing.Any]) -> None:
        deprecated_macos_show_window_title_in_menubar_alias('macos_show_window_title_in_menubar', val, ans)

    def send_text(self, val: str, ans: dict[str, typing.Any]) -> None:
        deprecated_send_text('send_text', val, ans)

    def adjust_line_height(self, val: str, ans: dict[str, typing.Any]) -> None:
        deprecated_adjust_line_height('adjust_line_height', val, ans)

    def adjust_column_width(self, val: str, ans: dict[str, typing.Any]) -> None:
        deprecated_adjust_line_height('adjust_column_width', val, ans)

    def adjust_baseline(self, val: str, ans: dict[str, typing.Any]) -> None:
        deprecated_adjust_line_height('adjust_baseline', val, ans)

    def scrollback_indicator_opacity(self, val: str, ans: dict[str, typing.Any]) -> None:
        deprecated_scrollback_indicator_opacity('scrollback_indicator_opacity', val, ans)

    def map(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k in parse_map(val):
            ans['map'].append(k)

    def mouse_map(self, val: str, ans: dict[str, typing.Any]) -> None:
        for k in parse_mouse_map(val):
            ans['mouse_map'].append(k)


def create_result_dict() -> dict[str, typing.Any]:
    return {
        'action_alias': {},
        'env': {},
        'exe_search_path': {},
        'filter_notification': {},
        'font_features': {},
        'kitten_alias': {},
        'menu_map': {},
        'modify_font': {},
        'narrow_symbols': {},
        'remote_control_password': {},
        'symbol_map': {},
        'watcher': {},
        'map': [],
        'mouse_map': [],
    }


actions: frozenset[str] = frozenset(('map', 'mouse_map'))


def merge_result_dicts(defaults: dict[str, typing.Any], vals: dict[str, typing.Any]) -> dict[str, typing.Any]:
    ans = {}
    for k, v in defaults.items():
        if isinstance(v, dict):
            ans[k] = merge_dicts(v, vals.get(k, {}))
        elif k in actions:
            ans[k] = v + vals.get(k, [])
        else:
            ans[k] = vals.get(k, v)
    return ans


parser = Parser()


def parse_conf_item(key: str, val: str, ans: dict[str, typing.Any]) -> bool:
    func = getattr(parser, key, None)
    if func is not None:
        func(val, ans)
        return True
    return False



#!/usr/bin/env python
# License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>

from collections.abc import Iterable
from typing import TYPE_CHECKING

from .base import MATCH_TAB_OPTION, ArgsType, Boss, PayloadGetType, PayloadType, RCOptions, RemoteCommand, ResponseType, UnknownLayout, Window

if TYPE_CHECKING:
    from kitty.cli_stub import GotoLayoutRCOptions as CLIOptions


def layout_names() -> Iterable[str]:
    from kitty.layout.interface import all_layouts
    return all_layouts.keys()


class GotoLayout(RemoteCommand):

    protocol_spec = __doc__ = '''
    layout+/str: The new layout name
    match/str: Which tab to change the layout of
    '''

    short_desc = 'Set the window layout'
    desc = (
        'Set the window layout in the specified tabs (or the active tab if not specified).'
        ' You can use special match value :code:`all` to set the layout in all tabs.'
        ' In case there are multiple layouts with the same name but different options,'
        ' specify the full layout definition or a unique prefix of the full definition.'
    )
    options_spec = MATCH_TAB_OPTION
    args = RemoteCommand.Args(
        spec='LAYOUT_NAME', count=1, json_field='layout',
        completion=RemoteCommand.CompletionSpec.from_string('type:keyword group:"Layout" kwds:' + ','.join(layout_names())),
        )

    def message_to_kitty(self, global_opts: RCOptions, opts: 'CLIOptions', args: ArgsType) -> PayloadType:
        if len(args) != 1:
            self.fatal('Exactly one layout must be specified')
        return {'layout': args[0], 'match': opts.match}

    def response_from_kitty(self, boss: Boss, window: Window | None, payload_get: PayloadGetType) -> ResponseType:
        tabs = self.tabs_for_match_payload(boss, window, payload_get)
        for tab in tabs:
            if tab:
                try:
                    tab.goto_layout(payload_get('layout'), raise_exception=True)
                except ValueError:
                    raise UnknownLayout('The layout {} is unknown or disabled or the name is ambiguous'.format(payload_get('layout')))
        return None


goto_layout = GotoLayout()



#!/usr/bin/env python
# License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>


from . import BaseTest


class TestTUI(BaseTest):

    def test_line_edit(self):
        from kittens.tui.line_edit import LineEdit
        le = LineEdit()
        le.on_text('abcd', False)
        self.ae(le.cursor_pos, 4)
        for i in range(5):
            self.assertTrue(le.left()) if i < 4 else self.assertFalse(le.left())
            self.ae(le.cursor_pos, max(0, 3 - i))
        self.ae(le.pending_bell, True)
        le.clear()
        le.on_text('abcd', False), le.home()
        self.ae(le.cursor_pos, 0)
        for i in range(5):
            self.assertTrue(le.right()) if i < 4 else self.assertFalse(le.right())
            self.ae(le.cursor_pos, min(4, i + 1))
        self.ae(le.pending_bell, True)
        le.clear()
        le.on_text('abcd', False)
        self.ae(le.current_input, 'abcd')
        self.ae(le.cursor_pos, 4)
        self.ae(le.split_at_cursor(), ('abcd', ''))
        le.backspace()
        self.ae(le.current_input, 'abc')
        self.ae(le.cursor_pos, 3)
        self.assertFalse(le.pending_bell)
        le.backspace(num=2)
        self.ae(le.current_input, 'a')
        self.ae(le.cursor_pos, 1)
        self.assertFalse(le.pending_bell)
        le.backspace(num=2)
        self.ae(le.current_input, '')
        self.ae(le.cursor_pos, 0)
        le.backspace()
        self.assertTrue(le.pending_bell)

    def test_multiprocessing_spawn(self):
        from kitty.multiprocessing import test_spawn
        test_spawn()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>

from collections.abc import Callable, Generator, Sequence
from functools import lru_cache
from itertools import repeat
from math import ceil, floor
from typing import Any

from kitty.borders import BorderColor
from kitty.types import Edges, NeighborsMap, WindowMapper
from kitty.typing_compat import WindowType
from kitty.window_list import WindowGroup, WindowList

from .base import BorderLine, Layout, LayoutData, LayoutDimension, ListOfWindows, layout_dimension, lgd
from .tall import neighbors_for_tall_window


@lru_cache
def calc_grid_size(n: int) -> tuple[int, int, int, int]:
    if n <= 5:
        ncols = 1 if n == 1 else 2
    else:
        for ncols in range(3, (n // 2) + 1):
            if ncols * ncols >= n:
                break
    nrows = n // ncols
    special_rows = n - (nrows * (ncols - 1))
    special_col = 0 if special_rows < nrows else ncols - 1
    return ncols, nrows, special_rows, special_col


class Grid(Layout):

    name: str = 'grid'
    no_minimal_window_borders = True

    def remove_all_biases(self) -> bool:
        self.biased_rows: dict[int, float] = {}
        self.biased_cols: dict[int, float] = {}
        return True

    def column_layout(
        self,
        num: int,
        bias: Sequence[float] | None = None,
    ) -> LayoutDimension:
        decoration_pairs = tuple(repeat((0, 0), num))
        return layout_dimension(lgd.central.left, lgd.central.width, lgd.cell_width, decoration_pairs, bias=bias, alignment=lgd.alignment_x)

    def row_layout(
        self,
        num: int,
        bias: Sequence[float] | None = None,
    ) -> LayoutDimension:
        decoration_pairs = tuple(repeat((0, 0), num))
        return layout_dimension(lgd.central.top, lgd.central.height, lgd.cell_height, decoration_pairs, bias=bias, alignment=lgd.alignment_y)

    def variable_layout(self, layout_func: Callable[..., LayoutDimension], num_windows: int, biased_map: dict[int, float]) -> LayoutDimension:
        return layout_func(num_windows, bias=biased_map if num_windows > 1 else None)

    def position_for_window_idx(self, idx: int, num_windows: int, ncols:int , nrows: int, special_rows: int, special_col: int) -> tuple[int, int]:
        row_num = col_num = 0

        def on_col_done(col_windows: list[int]) -> None:
            nonlocal col_num, row_num
            row_num = 0
            col_num += 1

        for window_idx, xl, yl in self.layout_windows(
                num_windows, nrows, ncols, special_rows, special_col, on_col_done):
            if idx == window_idx:
                return row_num, col_num
            row_num += 1
        return 0, 0

    def bias_slot(self, all_windows: WindowList, idx: int, fractional_bias: float, cell_increment_bias_h: float, cell_increment_bias_v: float) -> bool:
        num_windows = all_windows.num_groups
        ncols, nrows, special_rows, special_col = calc_grid_size(num_windows)
        row_num, col_num = self.position_for_window_idx(idx, num_windows, ncols, nrows, special_rows, special_col)
        if row_num == 0:
            b = self.biased_cols
            layout_func = self.column_layout
            bias_idx = col_num
            increment = cell_increment_bias_h
        else:
            b = self.biased_rows
            layout_func = self.row_layout
            bias_idx = row_num
            increment = cell_increment_bias_v
        before_layout = tuple(self.variable_layout(layout_func, num_windows, b))
        b[bias_idx] = increment
        return tuple(self.variable_layout(layout_func, num_windows, b)) == before_layout

    def apply_bias(self, idx: int, increment: float, all_windows: WindowList, is_horizontal: bool = True) -> bool:
        num_windows = all_windows.num_groups
        ncols, nrows, special_rows, special_col = calc_grid_size(num_windows)
        row_num, col_num = self.position_for_window_idx(idx, num_windows, ncols, nrows, special_rows, special_col)

        if is_horizontal:
            b = self.biased_cols
            if ncols < 2:
                return False
            bias_idx = col_num
            attr = 'biased_cols'

            def layout_func(windows: ListOfWindows, bias: Sequence[float] | None = None) -> LayoutDimension:
                return self.column_layout(num_windows, bias=bias)

        else:
            b = self.biased_rows
            if max(nrows, special_rows) < 2:
                return False
            bias_idx = row_num
            attr = 'biased_rows'

            def layout_func(windows: ListOfWindows, bias: Sequence[float] | None = None) -> LayoutDimension:
                return self.row_layout(num_windows, bias=bias)

        before_layout = tuple(self.variable_layout(layout_func, num_windows, b))
        candidate = b.copy()
        before = candidate.get(bias_idx, 0)
        candidate[bias_idx] = before + increment
        if before_layout == tuple(self.variable_layout(layout_func, num_windows, candidate)):
            return False
        setattr(self, attr, candidate)
        return True

    def layout_windows(
        self,
        num_windows: int,
        nrows: int, ncols: int,
        special_rows: int, special_col: int,
        on_col_done: Callable[[list[int]], None] = lambda col_windows: None
    ) -> Generator[tuple[int, LayoutData, LayoutData], None, None]:
        # Distribute windows top-to-bottom, left-to-right (i.e. in columns)
        xlayout = self.variable_layout(self.column_layout, ncols, self.biased_cols)
        yvals_normal = tuple(self.variable_layout(self.row_layout, nrows, self.biased_rows))
        yvals_special = yvals_normal if special_rows == nrows else tuple(self.variable_layout(self.row_layout, special_rows, self.biased_rows))
        pos = 0
        for col in range(ncols):
            rows = special_rows if col == special_col else nrows
            yls = yvals_special if col == special_col else yvals_normal
            xl = next(xlayout)
            col_windows = []
            for i, yl in enumerate(yls):
                window_idx = pos + i
                yield window_idx, xl, yl
                col_windows.append(window_idx)
            pos += rows
            on_col_done(col_windows)

    def do_layout(self, all_windows: WindowList) -> None:
        n = all_windows.num_groups
        if n == 1:
            self.layout_single_window_group(next(all_windows.iter_all_layoutable_groups()))
            return
        ncols, nrows, special_rows, special_col = calc_grid_size(n)
        groups = tuple(all_windows.iter_all_layoutable_groups())
        win_col_map: list[list[WindowGroup]] = []

        def on_col_done(col_windows: list[int]) -> None:
            col_windows_w = [groups[i] for i in col_windows]
            win_col_map.append(col_windows_w)

        def extents(ld: LayoutData) -> tuple[int, int]:
            start = ld.content_pos - ld.space_before
            size = ld.space_before + ld.space_after + ld.content_size
            return start, size

        def layout(ld: LayoutData, cell_length: int, before_dec: int, after_dec: int) -> LayoutData:
            start, size = extents(ld)
            space_needed_for_decorations = before_dec + after_dec
            content_size = size - space_needed_for_decorations
            number_of_cells = content_size // cell_length
            cell_area = number_of_cells * cell_length
            extra = content_size - cell_area
            if lgd.alignment_x == 0:  # center
                before_dec += extra // 2
            elif lgd.alignment_x > 0:  # end
                before_dec += extra
            return LayoutData(start + before_dec, number_of_cells, before_dec, size - cell_area - before_dec, cell_area)

        def position_window_in_grid_cell(window_idx: int, xl: LayoutData, yl: LayoutData) -> None:
            wg = groups[window_idx]
            edges = Edges(wg.decoration('left'), wg.decoration('top'), wg.decoration('right'), wg.decoration('bottom'))
            xl = layout(xl, lgd.cell_width, edges.left, edges.right)
            yl = layout(yl, lgd.cell_height, edges.top, edges.bottom)
            self.set_window_group_geometry(wg, xl, yl)

        for window_idx, xl, yl in self.layout_windows(
                n, nrows, ncols, special_rows, special_col, on_col_done):
            position_window_in_grid_cell(window_idx, xl, yl)

    def minimal_borders(self, all_windows: WindowList) -> Generator[BorderLine, None, None]:
        n = all_windows.num_groups
        if not lgd.draw_minimal_borders or n < 2:
            return
        needs_borders_map = all_windows.compute_needs_borders_map(lgd.draw_active_borders)
        ncols, nrows, special_rows, special_col = calc_grid_size(n)
        is_first_row: set[int] = set()
        is_last_row: set[int] = set()
        is_first_column: set[int] = set()
        is_last_column: set[int] = set()
        groups = tuple(all_windows.iter_all_layoutable_groups())
        bw = groups[0].effective_border()
        if not bw:
            return
        xl: LayoutData = LayoutData()
        yl: LayoutData = LayoutData()
        prev_col_windows: list[int] = []
        layout_data_map: dict[int, tuple[LayoutData, LayoutData]] = {}

        def on_col_done(col_windows: list[int]) -> None:
            nonlocal prev_col_windows, is_first_column
            if col_windows:
                is_first_row.add(groups[col_windows[0]].id)
                is_last_row.add(groups[col_windows[-1]].id)
            if not prev_col_windows:
                is_first_column = {groups[x].id for x in col_windows}
            prev_col_windows = col_windows

        all_groups_in_order: list[WindowGroup] = []
        for window_idx, xl, yl in self.layout_windows(n, nrows, ncols, special_rows, special_col, on_col_done):
            wg = groups[window_idx]
            all_groups_in_order.append(wg)
            layout_data_map[wg.id] = xl, yl
        is_last_column = {groups[x].id for x in prev_col_windows}
        active_group = all_windows.active_group

        def ends(yl: LayoutData) -> tuple[int, int]:
            return yl.content_pos - yl.space_before, yl.content_pos + yl.content_size + yl.space_after

        def borders_for_window(gid: int) -> Generator[Edges, None, None]:
            xl, yl = layout_data_map[gid]
            left, right = ends(xl)
            top, bottom = ends(yl)
            first_row, last_row = gid in is_first_row, gid in is_last_row
            first_column, last_column = gid in is_first_column, gid in is_last_column

            # Horizontal
            if not first_row:
                yield Edges(left, top, right, top + bw)
            if not last_row:
                yield Edges(left, bottom - bw, right, bottom)

            # Vertical
            if not first_column:
                yield Edges(left, top, left + bw, bottom)
            if not last_column:
                yield Edges(right - bw, top, right, bottom)

        for wg in all_groups_in_order:
            for edges in borders_for_window(wg.id):
                yield BorderLine(edges)
        for wg in all_groups_in_order:
            if needs_borders_map.get(wg.id):
                color = BorderColor.active if wg is active_group else BorderColor.bell
                for edges in borders_for_window(wg.id):
                    yield BorderLine(edges, color)

    def neighbors_for_window(self, window: WindowType, all_windows: WindowList) -> NeighborsMap:
        n = all_windows.num_groups
        if n < 4:
            return neighbors_for_tall_window(1, window, all_windows)

        wg = all_windows.group_for_window(window)
        assert wg is not None
        ncols, nrows, special_rows, special_col = calc_grid_size(n)
        blank_row: list[int | None] = [None for i in range(ncols)]
        matrix = tuple(blank_row[:] for j in range(max(nrows, special_rows)))
        wi = all_windows.iter_all_layoutable_groups()
        pos_map: dict[int, tuple[int, int]] = {}
        col_counts: list[int] = []
        for col in range(ncols):
            rows = special_rows if col == special_col else nrows
            for row in range(rows):
                w = next(wi)
                matrix[row][col] = wid = w.id
                pos_map[wid] = row, col
            col_counts.append(rows)
        row, col = pos_map[wg.id]

        def neighbors(row: int, col: int) -> list[int]:
            try:
                ans = matrix[row][col]
            except IndexError:
                ans = None
            return [] if ans is None else [ans]

        def side(row: int, col: int, delta: int) -> list[int]:
            neighbor_col = col + delta
            neighbor_nrows = col_counts[neighbor_col]
            nrows = col_counts[col]
            if neighbor_nrows == nrows:
                return neighbors(row, neighbor_col)

            start_row = floor(neighbor_nrows * row / nrows)
            end_row = ceil(neighbor_nrows * (row + 1) / nrows)
            xs = []
            for neighbor_row in range(start_row, end_row):
                xs.extend(neighbors(neighbor_row, neighbor_col))
            return xs

        ans: NeighborsMap = {}
        if row:
            ans['top'] = neighbors(row-1, col)
        if bottom := neighbors(row + 1, col):
            ans['bottom'] = bottom
        if col and (left := side(row, col, -1)):
            ans['left'] = left
        if col < ncols - 1:
            ans['right'] = side(row, col, 1)
        return ans

    def layout_state(self) -> dict[str, Any]:
        return {
            'biased_cols': self.biased_cols,
            'biased_rows': self.biased_rows
        }

    def set_layout_state(self, layout_state: dict[str, Any], map_group_id: WindowMapper) -> bool:
        self.biased_rows = {int(k): v for k, v in layout_state['biased_rows'].items()}
        self.biased_cols = {int(k): v for k, v in layout_state['biased_cols'].items()}
        return True



#!/usr/bin/env python
# License: GPLv3 Copyright: 2022, Kovid Goyal <kovid at kovidgoyal.net>

import os
import re
from typing import Dict, Iterator

tgt_pat = re.compile(r'^.. _(\S+?):$', re.MULTILINE)
title_pat = re.compile('^(.+)\n[-=^#*]{5,}$', re.MULTILINE)


def find_explicit_targets(text: str) -> Iterator[str]:
    for m in tgt_pat.finditer(text):
        yield m.group(1)


def find_page_title(text: str) -> str:
    for m in title_pat.finditer(text):
        return m.group(1)
    return ''


def main() -> Dict[str, Dict[str, str]]:
    refs = {}
    docs = {}
    base = os.path.dirname(os.path.abspath(__file__))
    for dirpath, dirnames, filenames in os.walk(base):
        if 'generated' in dirnames:
            dirnames.remove('generated')
        for f in filenames:
            if f.endswith('.rst'):
                with open(os.path.join(dirpath, f)) as stream:
                    raw = stream.read()
                href = os.path.relpath(stream.name, base).replace(os.sep, '/')
                href = href.rpartition('.')[0] + '/'
                docs[href.rstrip('/')] = find_page_title(raw)
                first_line = raw.lstrip('\n').partition('\n')[0]
                first_target_added = False
                for explicit_target in find_explicit_targets(raw):
                    # Shorten the reference link to the top of the page.
                    # Note that anchor links should still be used in HTML docs
                    # to allow jumping within the same page.
                    if not first_target_added:
                        first_target_added = True
                        if first_line.startswith(f'.. _{explicit_target}:'):
                            refs[explicit_target] = href
                            continue
                    refs[explicit_target] = href + f'#{explicit_target.replace("_", "-")}'
    return {'ref': refs, 'doc': docs}


if __name__ == '__main__':
    import json
    print(json.dumps(main(), indent=2))
